"""TcEx Framework Module"""
# standard library
from abc import ABC
from collections.abc import Generator
from typing import Any

# third-party
from pydantic import ValidationError
from requests.exceptions import ProxyError

# first-party
from tcex.api.tc.v3._gen._gen_abc import GenerateABC
from tcex.api.tc.v3._gen.model import FilterModel
from tcex.pleb.cached_property import cached_property
from tcex.util.render.render import Render
from tcex.util.string_operation import SnakeString


class GenerateFilterABC(GenerateABC, ABC):
    """Generate Models for Case Management Types"""

    def __init__(self, type_: SnakeString):
        """Initialize instance properties."""
        super().__init__(type_)

        # properties
        self.api_url = ''
        self.requirements = {
            'standard library': [
                'from enum import Enum',
            ],
            'third-party': [],
            'first-party': [
                'from tcex.api.tc.v3.api_endpoints import ApiEndpoints',
                'from tcex.api.tc.v3.filter_abc import FilterABC',
                'from tcex.api.tc.v3.tql.tql_type import TqlType',
            ],
            'first-party-forward-reference': [],
        }

    def _add_tql_imports(self):
        """Add TQL imports when required."""
        if 'from tcex.api.tc.v3.tql.tql import Tql' not in self.requirements['first-party']:
            self.requirements['first-party'].extend(
                [
                    'from tcex.api.tc.v3.tql.tql import Tql',
                    'from tcex.api.tc.v3.tql.tql_operator import TqlOperator',
                ]
            )

    @cached_property
    def _filter_contents(self) -> list[dict[str, Any]]:
        """Return defined API properties for the current object.

        Response:
        {
            "keyword": "analyticsScore",
            "name": "Analytics Score",
            "type": "Integer",
            "description": "The intel score of the artifact",
            "groupable": false,
            "targetable": true
        }
        """
        _properties = []
        try:
            r = self.session.options(f'{self.api_url}/tql', params={})
            # print(r.request.method, r.request.url, r.text)
            if r.ok:
                _properties = r.json().get('data', [])
        except (ConnectionError, ProxyError) as ex:
            Render.panel.failure(f'Failed getting types properties ({ex}).')

        return _properties

    @property
    def _filter_contents_updated(self) -> list[dict[str, Any]]:
        """Update the properties contents, fixing issues in core data."""
        filters = self._filter_contents

        # Provided invalid type "BigInteger" on tql options call and correcting it to String
        for filter_ in filters:
            if self.type_ == 'indicators':
                title = 'Indicator Keyword Type'
                if filter_['keyword'] == 'addressIpval' and filter_['type'] != 'String':
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix required.')
                    filter_['type'] = 'String'
                else:
                    self.messages.append(f'- [{self.type_}] - ({title}) - fix NOT required.')

            if self.type_ == 'cases' and 'description' in filter_:
                # fix misspelling in core data
                miss_map = {
                    'occured': 'occurred',
                    'Threatassess': 'ThreatAssess',
                }
                for c, w in miss_map.items():
                    if c in filter_['description']:
                        filter_['description'] = filter_['description'].replace(c, w)
        return sorted(filters, key=lambda i: i['keyword'])

    @cached_property
    def _filter_models(self) -> Generator[FilterModel, None, None]:
        for field_data in self._filter_contents_updated:
            try:
                yield FilterModel(**field_data)
            except ValidationError as ex:
                Render.panel.failure(f'Failed generating property model: data={field_data} ({ex}).')

    def _gen_code_generic_method(self, filter_data: FilterModel) -> list:
        """Return code for generic TQL filter methods."""
        keyword_description = self._format_description(
            arg=filter_data.keyword.snake_case(),
            description=filter_data.description,
            length=100,
            indent=' ' * 12,
        )
        _code = [
            (
                f'{self.i1}def {filter_data.keyword.snake_case()}'
                f'(self, operator: Enum, {filter_data.keyword.snake_case()}: '
                f'''{filter_data.extra.typing_type}):{filter_data.extra.comment}'''
            ),
            f'{self.i2}"""Filter {filter_data.name} based on **{filter_data.keyword}** keyword.',
            '',
            f'{self.i2}Args:',
            f'{self.i3}operator: The operator enum for the filter.',
            f'{self.i3}{keyword_description}',
            f'{self.i2}"""',
        ]
        if filter_data.type.lower() in ['date', 'datetime']:
            self.requirements['first-party'].extend(
                [
                    'from arrow import Arrow',
                    'from datetime import datetime',
                ]
            )
            _code.extend(
                [
                    f'''{self.i2}{filter_data.keyword.snake_case()} = self.util.any_to_datetime'''
                    f'''({filter_data.keyword.snake_case()}).strftime('%Y-%m-%d %H:%M:%S')'''
                ]
            )

        if 'list' in filter_data.extra.typing_type:
            _code.extend(
                [
                    (
                        f'''{self.i2}if isinstance({filter_data.keyword.snake_case()}, list) '''
                        '''and operator not in self.list_types:'''
                    ),
                    f'''{self.i3}raise RuntimeError('''
                    f'''{self.i5}'Operator must be CONTAINS, NOT_CONTAINS, IN\''''
                    f'''{self.i5}'or NOT_IN when filtering on a list of values.\''''
                    f'''{self.i4})''',
                    '',
                ]
            )
        _code.extend(
            [
                (
                    f'''{self.i2}self._tql.add_filter('{filter_data.keyword}', operator, '''
                    f'''{filter_data.keyword.snake_case()}, '''
                    f'''{filter_data.extra.tql_type})'''
                ),
                '',
            ]
        )
        return _code

    def _gen_code_has_artifact_method(self) -> list:
        """Return code for has_artifact TQL filter methods."""
        self._add_tql_imports()
        return [
            f'{self.i1}@property',
            f'{self.i1}def has_artifact(self):',
            f'{self.i2}"""Return **ArtifactFilter** for further filtering."""',
            f'{self.i2}# first-party',
            f'{self.i2}from tcex.api.tc.v3.artifacts.artifact_filter import ArtifactFilter',
            '',
            f'{self.i2}artifacts = ArtifactFilter(Tql())',
            (
                f'''{self.i2}self._tql.add_filter('hasArtifact', '''
                '''TqlOperator.EQ, artifacts, TqlType.SUB_QUERY)'''
            ),
            f'{self.i2}return artifacts',
            '',
        ]

    def _gen_code_has_case_method(self) -> list:
        """Return code for has_case TQL filter methods."""
        self._add_tql_imports()
        _code = [
            f'{self.i1}@property',
            f'{self.i1}def has_case(self):',
            f'{self.i2}"""Return **CaseFilter** for further filtering."""',
        ]
        if self.type_ != 'cases':
            _code.extend(
                [
                    f'{self.i2}# first-party',
                    f'{self.i2}from tcex.api.tc.v3.cases.case_filter import CaseFilter',
                    '',
                ]
            )
        _code.extend(
            [
                f'{self.i2}cases = CaseFilter(Tql())',
                (
                    f'''{self.i2}self._tql.add_filter('hasCase', '''
                    '''TqlOperator.EQ, cases, TqlType.SUB_QUERY)'''
                ),
                f'{self.i2}return cases',
                '',
            ]
        )
        return _code

    def _gen_code_has_group_method(self) -> list:
        """Return code for has_group TQL filter methods."""
        self._add_tql_imports()
        _code = [
            f'{self.i1}@property',
            f'{self.i1}def has_group(self):',
            f'{self.i2}"""Return **GroupFilter** for further filtering."""',
        ]
        if self.type_ != 'groups':
            _code.extend(
                [
                    f'{self.i2}# first-party',
                    f'{self.i2}from tcex.api.tc.v3.groups.group_filter import GroupFilter',
                    '',
                ]
            )
        _code.extend(
            [
                f'{self.i2}groups = GroupFilter(Tql())',
                (
                    f'''{self.i2}self._tql.add_filter('hasGroup', '''
                    '''TqlOperator.EQ, groups, TqlType.SUB_QUERY)'''
                ),
                f'{self.i2}return groups',
                '',
            ]
        )
        return _code

    def _gen_code_has_indicator_method(self) -> list:
        """Return code for has_indicator TQL filter methods."""
        self._add_tql_imports()
        _code = [
            f'{self.i1}@property',
            f'{self.i1}def has_indicator(self):',
            f'{self.i2}"""Return **IndicatorFilter** for further filtering."""',
        ]
        if self.type_ != 'indicators':
            _code.extend(
                [
                    f'{self.i2}# first-party',
                    (
                        f'{self.i2}from tcex.api.tc.v3.indicators.indicator_filter '
                        'import IndicatorFilter'
                    ),
                    '',
                ]
            )
        _code.extend(
            [
                f'{self.i2}indicators = IndicatorFilter(Tql())',
                (
                    f'''{self.i2}self._tql.add_filter('hasIndicator', '''
                    '''TqlOperator.EQ, indicators, TqlType.SUB_QUERY)'''
                ),
                f'{self.i2}return indicators',
                '',
            ]
        )
        return _code

    def _gen_code_has_note_method(self) -> list:
        """Return code for has_note TQL filter methods."""
        self._add_tql_imports()
        return [
            f'{self.i1}@property',
            f'{self.i1}def has_note(self):',
            f'{self.i2}"""Return **NoteFilter** for further filtering."""',
            f'{self.i2}# first-party',
            f'{self.i2}from tcex.api.tc.v3.notes.note_filter import NoteFilter',
            '',
            f'{self.i2}notes = NoteFilter(Tql())',
            (
                f'''{self.i2}self._tql.add_filter('hasNote', '''
                f'''TqlOperator.EQ, notes, TqlType.SUB_QUERY)'''
            ),
            f'{self.i2}return notes',
            '',
        ]

    def _gen_code_has_tag_method(self) -> list:
        """Return code for has_tag TQL filter methods."""
        self._add_tql_imports()
        return [
            f'{self.i1}@property',
            f'{self.i1}def has_tag(self):',
            f'{self.i2}"""Return **TagFilter** for further filtering."""',
            f'{self.i2}# first-party',
            f'{self.i2}from tcex.api.tc.v3.tags.tag_filter import TagFilter',
            '',
            f'{self.i2}tags = TagFilter(Tql())',
            (
                f'''{self.i2}self._tql.add_filter('hasTag', '''
                '''TqlOperator.EQ, tags, TqlType.SUB_QUERY)'''
            ),
            f'{self.i2}return tags',
            '',
        ]

    def _gen_code_has_task_method(self) -> list:
        """Return code for has_task TQL filter methods."""
        self._add_tql_imports()
        return [
            f'{self.i1}@property',
            f'{self.i1}def has_task(self):',
            f'{self.i2}"""Return **TaskFilter** for further filtering."""',
            f'{self.i2}# first-party',
            f'{self.i2}from tcex.api.tc.v3.tasks.task_filter import TaskFilter',
            '',
            f'{self.i2}tasks = TaskFilter(Tql())',
            (
                f'''{self.i2}self._tql.add_filter('hasTask', '''
                '''TqlOperator.EQ, tasks, TqlType.SUB_QUERY)'''
            ),
            f'{self.i2}return tasks',
            '',
        ]

    def _gen_code_has_security_label_method(self) -> list:
        """Return code for has_security_label TQL filter methods."""
        self._add_tql_imports()
        return [
            f'{self.i1}@property',
            f'{self.i1}def has_security_label(self):',
            f'{self.i2}"""Return **SecurityLabel** for further filtering."""',
            f'{self.i2}# first-party',
            (
                f'{self.i2}from tcex.api.tc.v3.security_labels.security_label_filter import '
                f'SecurityLabelFilter'
            ),
            '',
            f'{self.i2}security_labels = SecurityLabelFilter(Tql())',
            (
                f'''{self.i2}self._tql.add_filter('hasSecurityLabel', '''
                '''TqlOperator.EQ, security_labels, TqlType.SUB_QUERY)'''
            ),
            f'{self.i2}return security_labels',
            '',
        ]

    def _gen_code_has_victim_asset_method(self) -> list:
        """Return code for has_victim_asset TQL filter methods."""
        self._add_tql_imports()
        _code = [
            f'{self.i1}@property',
            f'{self.i1}def has_victim_asset(self):',
            f'{self.i2}"""Return **VictimAssetFilter** for further filtering."""',
        ]
        if self.type_ != 'victim_assets':
            _code.extend(
                [
                    f'{self.i2}# first-party',
                    (
                        f'{self.i2}from tcex.api.tc.v3.victim_assets.victim_asset_filter import '
                        f'VictimAssetFilter'
                    ),
                    '',
                ]
            )
        _code.extend(
            [
                f'{self.i2}victim_assets = VictimAssetFilter(Tql())',
                (
                    f'''{self.i2}self._tql.add_filter('hasVictimAsset', '''
                    '''TqlOperator.EQ, victim_assets, TqlType.SUB_QUERY)'''
                ),
                f'{self.i2}return victim_assets',
                '',
            ]
        )
        return _code

    def _gen_code_has_victim_method(self) -> list:
        """Return code for has_victim TQL filter methods."""
        self._add_tql_imports()
        _code = [
            f'{self.i1}@property',
            f'{self.i1}def has_victim(self):',
            f'{self.i2}"""Return **VictimFilter** for further filtering."""',
        ]
        if self.type_ != 'victims':
            _code.extend(
                [
                    f'{self.i2}# first-party',
                    f'{self.i2}from tcex.api.tc.v3.victims.victim_filter import VictimFilter',
                    '',
                ]
            )
        _code.extend(
            [
                f'{self.i2}victims = VictimFilter(Tql())',
                (
                    f'''{self.i2}self._tql.add_filter('hasVictim', '''
                    '''TqlOperator.EQ, victims, TqlType.SUB_QUERY)'''
                ),
                f'{self.i2}return victims',
                '',
            ]
        )
        return _code

    def _gen_code_has_attribute_method(self) -> list:
        """Return code for has_attribute TQL filter methods."""
        self._add_tql_imports()

        # hasAttribute filter is present on groups/indicators/ and victims.
        filter_type = 'GroupAttributeFilter'
        import_path = 'group_attributes.group_attribute_filter'
        if self.type_.lower() == 'indicators':
            filter_type = 'IndicatorAttributeFilter'
            import_path = 'indicator_attributes.indicator_attribute_filter'
        elif self.type_.lower() == 'victims':
            filter_type = 'VictimAttributeFilter'
            import_path = 'victim_attributes.victim_attribute_filter'

        return [
            f'{self.i1}@property',
            f'{self.i1}def has_attribute(self):',
            f'{self.i2}"""Return **{filter_type}** for further filtering."""',
            f'{self.i2}# first-party',
            f'{self.i2}from tcex.api.tc.v3.{import_path} import {filter_type}',
            '',
            f'{self.i2}attributes = {filter_type}(Tql())',
            (
                f'''{self.i2}self._tql.add_filter('hasAttribute', '''
                '''TqlOperator.EQ, attributes, TqlType.SUB_QUERY)'''
            ),
            f'{self.i2}return attributes',
            '',
        ]

    def gen_api_endpoint_method(self) -> str:
        r"""Generate private class method/property.

        @property
        def _api_endpoint(self) -> str:
            \"\"\"Return the API endpoint.\"\"\"
            return ApiEndpoints.ARTIFACTS.value
        """
        _method = [
            f'{self.i1}@property',
            f'{self.i1}def _api_endpoint(self) -> str:',
            f'{self.i2}"""Return the API endpoint."""',
            f'{self.i2}return ApiEndpoints.{self.type_.upper()}.value',
        ]
        _method.append('')
        return '\n'.join(_method)

    def gen_class(self) -> str:
        """Generate doc string."""
        return '\n'.join(
            [
                f'\nclass {self.type_.singular().pascal_case()}Filter(FilterABC):',
                f'{self.i1}"""Filter Object for {self.type_.plural().pascal_case()}"""',
                '',
                '',
            ]
        )

    def gen_class_methods(self):
        """Return a Filter Class current object."""
        _filter_class = []

        # added _api_endpoint method
        _filter_class.append(self.gen_api_endpoint_method())

        for f in self._filter_models:
            if f.keyword.snake_case() == 'has_artifact':
                _filter_class.extend(self._gen_code_has_artifact_method())
            elif f.keyword.snake_case() == 'has_case':
                _filter_class.extend(self._gen_code_has_case_method())
            elif f.keyword.snake_case() == 'has_group':
                _filter_class.extend(self._gen_code_has_group_method())
            elif f.keyword.snake_case() == 'has_indicator':
                _filter_class.extend(self._gen_code_has_indicator_method())
            elif f.keyword.snake_case() == 'has_note':
                _filter_class.extend(self._gen_code_has_note_method())
            elif f.keyword.snake_case() == 'has_tag':
                _filter_class.extend(self._gen_code_has_tag_method())
            elif f.keyword.snake_case() == 'has_task':
                _filter_class.extend(self._gen_code_has_task_method())
            elif f.keyword.snake_case() == 'has_attribute':
                _filter_class.extend(self._gen_code_has_attribute_method())
            elif f.keyword.snake_case() == 'has_security_label':
                _filter_class.extend(self._gen_code_has_security_label_method())
            elif f.keyword.snake_case() == 'has_victim':
                _filter_class.extend(self._gen_code_has_victim_method())
            elif f.keyword.snake_case() == 'has_victim_asset':
                _filter_class.extend(self._gen_code_has_victim_asset_method())
            else:
                _filter_class.extend(self._gen_code_generic_method(f))

        return '\n'.join(_filter_class)

    def gen_doc_string(self) -> str:
        """Generate doc string."""
        # return f'"""{self.type_.singular().title()} TQL Filter"""\n'
        return '"""TcEx Framework Module"""\n'



"""TcEx Framework Module"""
# standard library
import operator
from collections.abc import Callable
from typing import Any

# third-party
from pydantic import BaseModel, validator
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.input.field_type.exception import InvalidEmptyValue, InvalidInput, InvalidType


def always_array(
    allow_empty: bool = True,
    include_empty: bool = False,
    include_null: bool = False,
    split_csv: bool = False,
) -> Callable[[Any, ModelField], list[Any]]:
    """Return customized validator that always returns a list.

    Args:
        allow_empty: If True, return empty list if input is empty.
        include_empty: If True, will wrap empty string in list. This does not
            affect list with existing empty strings.
        include_null: If True, will wrap null value in list. This does not
            affect list with existing null values.
        split_csv: if True and input value is a string, then string will be split on comma. No
        further processing is done on result of splitting on comma
    """

    def _always_array(value: Any, field: ModelField) -> list[Any]:
        """Return validator."""

        if split_csv and isinstance(value, str) and value:
            value = value.split(',')

        if include_empty is True and value == '':
            value = [value]

        if include_null is True and value is None:
            value = [value]

        if value in (None, ''):
            value = []
        elif not isinstance(value, list):
            value = [value]

        if allow_empty is False and value == []:
            raise InvalidEmptyValue(field_name=field.name)

        return value

    return _always_array


def conditional_required(rules: list[dict[str, str]]) -> Any:
    """Return customized validator that validates conditional required fields.

    Example Usage:

    If the severity field is set to 'High' the action field is required.

    MyModel(BaseModel):
        severity: str  # Low, Medium, High
        action: str | None

        _conditional_required = validator('action', allow_reuse=True, always=True, pre=True)(
            conditional_required(
                rules=[{'field': 'severity', 'operation': 'eq', 'value': 'High'}]
            )
        )
    """

    def is_in(input_: str, value: list) -> bool:
        """Return True if input is in value list."""
        return input_ in value

    def not_in(input_: str, value: list) -> bool:
        """Return True if input is NOT in value list."""
        return input_ not in value

    def get_operator(op):
        """Get the corresponding operator"""

        operators = {
            'eq': operator.eq,
            'in': is_in,
            'ne': operator.ne,
            'ni': not_in,
        }
        return operators.get(op, operator.eq)

    def _conditional_required(value: str, field: ModelField, values: dict[str, Any]):
        """Return validator."""
        for rule in rules or []:
            # the conditional field must be set above the conditionally required field
            conditional_field = values.get(rule['field'])
            conditional_operator = get_operator(rule.get('op', 'eq'))
            conditional_value = rule['value']

            if conditional_operator(conditional_field, conditional_value) is True and not value:
                raise InvalidInput(field_name=field.name, error='input is conditionally required.')

        return value

    return _conditional_required


def entity_input(
    allow_empty: bool = False,
    only_field: str | None = None,
    type_filter: list[str] | None = None,
) -> str | list[Any]:
    """Return customized validator.

    allow_empty:

    If this is False, an emptiness check is performed on the final value that is to be returned
    just before returning it. If the return value is either an empty string or an empty list, then
    an error is raised. If allow_empty is True, then the final value is not guaranteed not to be
    an empty string or empty list. Note: this occurs after the only_value and type_filter logic.
    If the value to be returned is a list that is not empty, no checks are performed on the list
    members.

    only_field:

    Valid inputs for this field are [`value`, `id`].

    When accepting String and TCEntity as PB data types, it is often easier to always work with
    the value or id of the TCEntity. This validator will ensure that only the value or id of the
    TCEntities is returned.

    type_filter:

    Filter TCEntities for a particular type. This parameter expects a list of strings with each
    string being a target TCEntity type to filter for. This parameter only has an effect when this
    method is working with TCEntities. When working with Strings, the values are not verified to be
    of any particular TCEntity type and are returned as is. Note: if target type list contains
    more than one type and only_value is True, then an array of values of mixed types could occur.

    Example Usage:

    MyModel(BaseModel):
        ip_address: AddressEntity | ip_address(strip_port=True)

        _entity_field = validator('ip_address', allow_reuse=True)(entity_field(only_value=True))
    """

    def _entity_input(
        value: BaseModel | list[BaseModel] | str | list[str], field: ModelField
    ) -> list[str] | str:
        """Return value from String or TCEntity."""

        def _get_value(value: str | BaseModel) -> BaseModel | str | None:
            """Return value"""
            if isinstance(value, BaseModel):
                if isinstance(type_filter, list) and value.type not in type_filter:  # type: ignore
                    return None

                if only_field is None:
                    return value
                if only_field.lower() == 'value':
                    return value.value  # type: ignore
                if only_field.lower() == 'id':
                    return value.id  # type: ignore
                raise InvalidInput(
                    field_name=field.name, error=f'Only Field {only_field} is not allowed.'
                )

            return value

        if isinstance(value, list):
            _values = []
            for v in value:
                _value = _get_value(v)
                if _value is not None:
                    _values.append(_value)
            value = _values
        else:
            # TODO: [high] fix this type ignore
            value = _get_value(value)  # type: ignore

        if field.allow_none is False and value is None:
            raise InvalidInput(field_name=field.name, error='None value is not allowed.')

        if allow_empty is False and value in ['', []]:
            raise InvalidInput(field_name=field.name, error='Empty value is not allowed.')

        # TODO: [high] fix this type ignore
        return value  # type: ignore

    # TODO: [high] fix this type ignore
    return _entity_input  # type: ignore


def modify_advanced_settings(input_name) -> Any:
    """Return validator that parses an advanced settings string and returns a dictionary

    :param input_name: the name of the input (within the app model) that will receive the
    pipe-delimited advanced settings string. In other words, the name of the input field
    that this validator should act on to parse the pipe-delimited string into a dictionary
    """

    def _modify_advanced_settings(value: Any, field: ModelField) -> dict[str, str]:
        """Return validator."""
        settings = {}

        if value is None:
            if field.allow_none:
                return value
            raise InvalidInput(
                field_name=field.name,
                error='Value for field is None (null) and field is not Optional',
            )

        if not isinstance(value, str):
            raise InvalidType(
                field_name=field.name, expected_types='(str)', provided_type=type(value)
            )

        entries = value.split('|')

        for entry in entries:
            key_value = entry.split('=', maxsplit=1)

            if len(key_value) == 1:
                continue

            key, value = key_value
            key = key.strip().lower()
            if not key:
                continue

            if key in settings:
                raise InvalidInput(
                    field_name=field.name,
                    error=f'Duplicate key "{key}" defined in Advanced Settings input.',
                )

            settings[key] = value

        return settings

    return validator(input_name, pre=True, allow_reuse=True)(_modify_advanced_settings)



"""TcEx Framework Module"""
# standard library
import ipaddress
import re
from typing import Any

from .aes_operation import AesOperation
from .datetime_operation import DatetimeOperation
from .string_operation import StringOperation
from .variable import Variable


class Util(AesOperation, DatetimeOperation, StringOperation, Variable):
    """TcEx Utilities Class"""

    @staticmethod
    def flatten_list(lst: list[Any]) -> list[Any]:
        """Flatten a list

        Will work for lists of lists to arbitrary depth
        and for lists with a mix of lists and single values
        """
        flat_list = []
        for sublist in lst:
            if isinstance(sublist, list):
                for item in Util.flatten_list(sublist):
                    flat_list.append(item)
            else:
                flat_list.append(sublist)
        return flat_list

    @staticmethod
    def is_cidr(possible_cidr_range: str) -> bool:
        """Return True if the provided value is a valid CIDR block."""
        try:
            ipaddress.ip_address(possible_cidr_range)
        except ValueError:
            try:
                ipaddress.ip_interface(possible_cidr_range)
            except Exception:
                return False
            return True
        return False

    @staticmethod
    def is_ip(possible_ip: str) -> bool:
        """Return True if the provided value is a valid IP address."""
        try:
            ipaddress.ip_address(possible_ip)
        except ValueError:
            return False
        return True

    @staticmethod
    def printable_cred(
        cred: str,
        visible: int = 1,
        mask_char: str = '*',
        mask_char_count: int = 4,
    ) -> str:
        """Return a printable (masked) version of the provided credential.

        Args:
            cred: The cred to print.
            visible: The number of characters at the beginning and ending of the cred to not mask.
            mask_char: The character to use in the mask.
            mask_char_count: How many mask character to insert (obscure cred length).
        """
        visible = max(visible, 1)
        if isinstance(cred, str):
            mask_char = mask_char or '*'
            if cred is not None and len(cred) >= visible * 2:
                cred = f'{cred[:visible]}{mask_char * mask_char_count}{cred[-visible:]}'
        return cred

    @staticmethod
    def remove_none(dict_: dict[Any, Any | None]) -> dict[Any, Any]:
        """Remove any mapping from a single level dict with a None value."""
        return {k: v for k, v in dict_.items() if v is not None}

    @staticmethod
    def standardize_asn(asn: str) -> str:
        """Return the ASN formatted for ThreatConnect."""
        numbers = re.findall('[0-9]+', asn)
        if len(numbers) == 1:
            asn = f'ASN{numbers[0]}'
        return asn



"""TcEx Framework Module"""
# standard library
from datetime import datetime
from enum import Enum

# third-party
from arrow import Arrow

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class WorkflowEventFilter(FilterABC):
    """Filter Object for WorkflowEvents"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.WORKFLOW_EVENTS.value

    def case_id(self, operator: Enum, case_id: int | list):
        """Filter Case ID based on **caseId** keyword.

        Args:
            operator: The operator enum for the filter.
            case_id: The ID of the case this event is associated with.
        """
        if isinstance(case_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('caseId', operator, case_id, TqlType.INTEGER)

    def date_added(self, operator: Enum, date_added: Arrow | datetime | int | str):
        """Filter Date Added based on **dateAdded** keyword.

        Args:
            operator: The operator enum for the filter.
            date_added: The date the event was added.
        """
        date_added = self.util.any_to_datetime(date_added).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('dateAdded', operator, date_added, TqlType.STRING)

    def deleted(self, operator: Enum, deleted: bool):
        """Filter Deleted based on **deleted** keyword.

        Args:
            operator: The operator enum for the filter.
            deleted: The deletion status of the event.
        """
        self._tql.add_filter('deleted', operator, deleted, TqlType.BOOLEAN)

    def deleted_reason(self, operator: Enum, deleted_reason: list | str):
        """Filter Deleted Reason based on **deletedReason** keyword.

        Args:
            operator: The operator enum for the filter.
            deleted_reason: The reason the event was deleted.
        """
        if isinstance(deleted_reason, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('deletedReason', operator, deleted_reason, TqlType.STRING)

    def event_date(self, operator: Enum, event_date: Arrow | datetime | int | str):
        """Filter Event Date based on **eventDate** keyword.

        Args:
            operator: The operator enum for the filter.
            event_date: The date the event occurred.
        """
        event_date = self.util.any_to_datetime(event_date).strftime('%Y-%m-%d %H:%M:%S')
        self._tql.add_filter('eventDate', operator, event_date, TqlType.STRING)

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the event.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def link(self, operator: Enum, link: list | str):
        """Filter Link based on **link** keyword.

        Args:
            operator: The operator enum for the filter.
            link: The item this event pertains to, in format <type>:<id>.
        """
        if isinstance(link, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('link', operator, link, TqlType.STRING)

    def summary(self, operator: Enum, summary: list | str):
        """Filter Summary based on **summary** keyword.

        Args:
            operator: The operator enum for the filter.
            summary: Text of the event.
        """
        if isinstance(summary, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('summary', operator, summary, TqlType.STRING)

    def system_generated(self, operator: Enum, system_generated: bool):
        """Filter System Generated based on **systemGenerated** keyword.

        Args:
            operator: The operator enum for the filter.
            system_generated: Flag determining if this event was created automatically by the
                system.
        """
        self._tql.add_filter('systemGenerated', operator, system_generated, TqlType.BOOLEAN)

    def user_name(self, operator: Enum, user_name: list | str):
        """Filter User Name based on **userName** keyword.

        Args:
            operator: The operator enum for the filter.
            user_name: The username associated with the event.
        """
        if isinstance(user_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('userName', operator, user_name, TqlType.STRING)



"""TcEx Framework Module"""
# pylint: disable=no-member,no-self-argument,wrong-import-position
# standard library
from datetime import datetime

# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class CaseModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='Case Model',
    validate_assignment=True,
):
    """Case Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(True)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    artifacts: 'ArtifactsModel' = Field(
        None,
        description='A list of Artifacts corresponding to the Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='artifacts',
    )
    assignee: 'AssigneeModel' = Field(
        None,
        description='The user or group Assignee object for the Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='assignee',
    )
    associated_cases: 'CasesModel' = Field(
        None,
        description='A list of Cases associated with this Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedCases',
    )
    associated_groups: 'GroupsModel' = Field(
        None,
        description='A list of Groups associated with this Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedGroups',
    )
    associated_indicators: 'IndicatorsModel' = Field(
        None,
        description='A list of Indicators associated with this Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='associatedIndicators',
    )
    attributes: 'CaseAttributesModel' = Field(
        None,
        description='A list of Attributes corresponding to the Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='attributes',
    )
    case_close_time: datetime | None = Field(
        None,
        description='The date and time that the Case was closed.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='caseCloseTime',
    )
    case_close_user: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The user that closed the Case.',
        read_only=True,
        title='caseCloseUser',
    )
    case_detection_time: datetime | None = Field(
        None,
        description='The date and time that ends the user initiated Case duration.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='caseDetectionTime',
    )
    case_detection_user: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The user that stopped the clock on Case duration.',
        read_only=True,
        title='caseDetectionUser',
    )
    case_occurrence_time: datetime | None = Field(
        None,
        description='The date and time that starts the user initiated Case duration.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='caseOccurrenceTime',
    )
    case_occurrence_user: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The user that started the clock on Case duration.',
        read_only=True,
        title='caseOccurrenceUser',
    )
    case_open_time: datetime | None = Field(
        None,
        description='The date and time that the Case was first opened.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='caseOpenTime',
    )
    case_open_user: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The user that opened the Case.',
        read_only=True,
        title='caseOpenUser',
    )
    created_by: 'UserModel' = Field(
        None,
        allow_mutation=False,
        description='The **created by** for the Case.',
        read_only=True,
        title='createdBy',
    )
    date_added: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the Case was first created.',
        read_only=True,
        title='dateAdded',
    )
    description: str | None = Field(
        None,
        description='The description of the Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='description',
    )
    id: int | None = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    last_updated: datetime | None = Field(
        None,
        allow_mutation=False,
        description='The date and time that the Case was last updated.',
        read_only=True,
        title='lastUpdated',
    )
    name: str | None = Field(
        None,
        description='The name of the Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='name',
    )
    notes: 'NotesModel' = Field(
        None,
        description='A list of Notes corresponding to the Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='notes',
    )
    owner: str | None = Field(
        None,
        allow_mutation=False,
        description='The name of the Owner of the Case.',
        read_only=True,
        title='owner',
    )
    owner_id: int | None = Field(
        None,
        allow_mutation=False,
        description='The id of the Owner of the Case.',
        read_only=True,
        title='ownerId',
    )
    related: 'CasesModel' = Field(
        None,
        allow_mutation=False,
        description='The **related** for the Case.',
        read_only=True,
        title='related',
    )
    resolution: str | None = Field(
        None,
        description='The Case resolution.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='resolution',
    )
    severity: str | None = Field(
        None,
        description='The Case severity.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='severity',
    )
    status: str | None = Field(
        None,
        description='The Case status.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='status',
    )
    tags: 'TagsModel' = Field(
        None,
        description=(
            'A list of Tags corresponding to the Case (NOTE: Setting this parameter will replace '
            'any existing tag(s) with the one(s) specified).'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='tags',
    )
    tasks: 'TasksModel' = Field(
        None,
        description='A list of Tasks corresponding to the Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='tasks',
    )
    user_access: 'UsersModel' = Field(
        None,
        description=(
            'A list of Users that, when defined, are the only ones allowed to view or edit the '
            'Case.'
        ),
        methods=['POST', 'PUT'],
        read_only=False,
        title='userAccess',
    )
    workflow_events: 'WorkflowEventsModel' = Field(
        None,
        description='A list of workflowEvents (timeline) corresponding to the Case.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='workflowEvents',
    )
    workflow_template: 'WorkflowTemplateModel' = Field(
        None,
        description='The Template that the Case is populated by.',
        methods=['POST', 'PUT'],
        read_only=False,
        title='workflowTemplate',
    )
    xid: str | None = Field(
        None,
        description='The **xid** for the Case.',
        methods=['POST'],
        read_only=False,
        title='xid',
    )

    @validator('artifacts', always=True, pre=True)
    def _validate_artifacts(cls, v):
        if not v:
            return ArtifactsModel()  # type: ignore
        return v

    @validator('assignee', always=True, pre=True)
    def _validate_assignee(cls, v):
        if not v:
            return AssigneeModel()  # type: ignore
        return v

    @validator('attributes', always=True, pre=True)
    def _validate_case_attributes(cls, v):
        if not v:
            return CaseAttributesModel()  # type: ignore
        return v

    @validator('associated_cases', 'related', always=True, pre=True)
    def _validate_cases(cls, v):
        if not v:
            return CasesModel()  # type: ignore
        return v

    @validator('associated_groups', always=True, pre=True)
    def _validate_groups(cls, v):
        if not v:
            return GroupsModel()  # type: ignore
        return v

    @validator('associated_indicators', always=True, pre=True)
    def _validate_indicators(cls, v):
        if not v:
            return IndicatorsModel()  # type: ignore
        return v

    @validator('notes', always=True, pre=True)
    def _validate_notes(cls, v):
        if not v:
            return NotesModel()  # type: ignore
        return v

    @validator('tags', always=True, pre=True)
    def _validate_tags(cls, v):
        if not v:
            return TagsModel()  # type: ignore
        return v

    @validator('tasks', always=True, pre=True)
    def _validate_tasks(cls, v):
        if not v:
            return TasksModel()  # type: ignore
        return v

    @validator(
        'case_close_user',
        'case_detection_user',
        'case_occurrence_user',
        'case_open_user',
        'created_by',
        always=True,
        pre=True,
    )
    def _validate_user(cls, v):
        if not v:
            return UserModel()  # type: ignore
        return v

    @validator('user_access', always=True, pre=True)
    def _validate_users(cls, v):
        if not v:
            return UsersModel()  # type: ignore
        return v

    @validator('workflow_events', always=True, pre=True)
    def _validate_workflow_events(cls, v):
        if not v:
            return WorkflowEventsModel()  # type: ignore
        return v

    @validator('workflow_template', always=True, pre=True)
    def _validate_workflow_template(cls, v):
        if not v:
            return WorkflowTemplateModel()  # type: ignore
        return v


class CaseDataModel(
    BaseModel,
    title='Case Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Cases Data Model"""

    data: list[CaseModel] | None = Field(
        [],
        description='The data for the Cases.',
        methods=['POST', 'PUT'],
        title='data',
    )


class CasesModel(
    BaseModel,
    title='Cases Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """Cases Model"""

    _mode_support = PrivateAttr(True)

    data: list[CaseModel] | None = Field(
        [],
        description='The data for the Cases.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.artifacts.artifact_model import ArtifactsModel
from tcex.api.tc.v3.case_attributes.case_attribute_model import CaseAttributesModel
from tcex.api.tc.v3.groups.group_model import GroupsModel
from tcex.api.tc.v3.indicators.indicator_model import IndicatorsModel
from tcex.api.tc.v3.notes.note_model import NotesModel
from tcex.api.tc.v3.security.assignee_model import AssigneeModel
from tcex.api.tc.v3.security.users.user_model import UserModel, UsersModel
from tcex.api.tc.v3.tags.tag_model import TagsModel
from tcex.api.tc.v3.tasks.task_model import TasksModel
from tcex.api.tc.v3.workflow_events.workflow_event_model import WorkflowEventsModel
from tcex.api.tc.v3.workflow_templates.workflow_template_model import WorkflowTemplateModel

# add forward references
CaseDataModel.update_forward_refs()
CaseModel.update_forward_refs()
CasesModel.update_forward_refs()



"""TcEx Framework Module"""
# standard library
import time
from collections.abc import Callable

from ...input.field_type.sensitive import Sensitive  # type: ignore # pylint: disable=import-error
from .hmac_auth import HmacAuth
from .token_auth import TokenAuth


class TcAuth(HmacAuth, TokenAuth):
    """ThreatConnect Token Authorization"""

    def __init__(
        self,
        tc_api_access_id: str | None = None,
        tc_api_secret_key: Sensitive | None = None,
        tc_token: Callable | str | Sensitive | None = None,
    ):
        """Initialize instance properties."""
        if tc_api_access_id is not None and tc_api_secret_key is not None:
            HmacAuth.__init__(self, tc_api_access_id, tc_api_secret_key)
            self.auth_type = 'hmac'
        elif tc_token is not None:
            TokenAuth.__init__(self, tc_token)
            self.auth_type = 'token'
        else:  # pragma: no cover
            raise RuntimeError('No valid ThreatConnect API credentials provided.')

    def __call__(self, r):
        """Add the authorization headers to the request."""
        timestamp = int(time.time())
        if self.auth_type == 'hmac':
            r.headers['Authorization'] = self._hmac_header(r, timestamp)
        elif self.auth_type == 'token':
            r.headers['Authorization'] = self._token_header()

        # Add required headers to auth.
        r.headers['Timestamp'] = str(timestamp)
        return r



"""TcEx Framework Module"""
# standard library
import ipaddress
from collections.abc import Generator

# third-party
from pydantic.fields import ModelField  # TYPE-CHECKING

# first-party
from tcex.input.field_type.exception import InvalidInput


class IpAddress(str):
    """String Field Type"""

    strip_port: bool = False

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Run validators / modifiers on input."""
        yield cls.validate_strip_port
        yield cls.validate_ipaddress

    @classmethod
    def validate_strip_port(cls, value: str, field: ModelField) -> str:
        """Modify value when requested."""
        if cls.strip_port is True:
            if ':' in value:
                _value, port = value.rsplit(':', 1)

                try:
                    if int(port) > 65535 or int(port) < 0:
                        raise ValueError()
                except ValueError:
                    raise InvalidInput(field.name, f'Invalid IP Address provided ({value}).')

                value = _value

        return value

    @classmethod
    def validate_ipaddress(cls, value: str, field: ModelField) -> str:
        """Raise exception if value is not a String type."""
        try:
            ipaddress.ip_address(value)
        except ValueError as ex:
            raise InvalidInput(field.name, f'Invalid IP Address provided ({value}).') from ex

        return value


def ip_address(
    strip_port: bool = False,
) -> type[IpAddress]:
    """Return configured instance of String."""
    namespace = {
        'strip_port': strip_port,
    }
    return type('ConstrainedIpAddress', (IpAddress,), namespace)



"""TcEx Framework Module"""
# pylint: disable=no-self-argument

# third-party
from pydantic import BaseModel, Field, validator
from semantic_version import Version

__all__ = ['JobJsonModel']


def snake_to_camel(snake_string: str) -> str:
    """Convert snake_case to camelCase"""
    components = snake_string.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


class ParamsModel(BaseModel):
    """Model definition for job.json.params"""

    default: bool | str | None
    encrypt: bool = False
    name: str
    prevent_updates: bool = False

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        # without smart_union the default value as a string of "0" would be converted to a bool
        smart_union = True
        validate_assignment = True


class JobJsonCommonModel(BaseModel):
    """Model definition for common field in job.json."""

    allow_on_demand: bool = Field(
        ...,
        description='If true, the job can be run on demand.',
    )
    enable_notifications: bool = Field(
        ..., description='Enables pass/fail notification for this job.'
    )
    job_name: str
    notify_email: str = Field(
        ...,
        description='Email address to send notification to.',
    )
    notify_include_log_files: bool = Field(
        ...,
        description='If true, the job log files will be included in the notification email.',
    )
    notify_on_complete: bool = Field(
        ...,
        description='If true, a notification will be sent when the job completes.',
    )
    notify_on_failure: bool = Field(
        ...,
        description='If true, a notification will be sent when the job fails.',
    )
    notify_on_partial_failure: bool = Field(
        ...,
        description=(
            'If true, a notification will be sent when the job completes with partial success.'
        ),
    )
    params: list[ParamsModel]
    publish_auth: bool = Field(
        ...,
        description='If true, the job will publish the authentication token.',
    )
    schedule_cron_format: str
    schedule_start_date: int
    schedule_type: str

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        arbitrary_types_allowed = True
        validate_assignment = True


class JobJsonModel(JobJsonCommonModel):
    """Model definition for job.json configuration file"""

    program_name: str
    program_version: str

    @validator('program_version')
    def version(cls, v):
        """Return a version object for "version" fields."""
        if v is not None:
            return Version(v)
        return v  # pragma: no cover

    class Config:
        """DataModel Config"""

        alias_generator = snake_to_camel
        arbitrary_types_allowed = True
        json_encoders = {Version: lambda v: str(v)}  # pylint: disable=unnecessary-lambda
        validate_assignment = True



"""TcEx Framework Module"""
# standard library
from collections.abc import Generator
from typing import Any

# third-party
import arrow
from pydantic.fields import ModelField

# first-party
from tcex.input.field_type.exception import InvalidInput
from tcex.util import Util


class DateTime(arrow.Arrow):
    """Datetime (Arrow) Field Type"""

    @classmethod
    def __get_validators__(cls) -> Generator:
        """Define one or more validators for Pydantic custom type."""
        yield cls._validate

    @classmethod
    def _validate(cls, value: Any, field: ModelField) -> arrow.Arrow:
        """Pydantic validate method."""
        try:
            return Util.any_to_datetime(value)
        except RuntimeError as ex:
            raise InvalidInput(field_name=field.name, error=str(ex)) from ex



"""TcEx Framework Module"""
# third-party
from pydantic import BaseModel, Field


class BatchModel(BaseModel):
    """Batch Model

    Supported for the following runtimeLevel:
    * ApiService
    * Playbook
    * Organization
    * WebhookTriggerService
    * TriggerService
    """

    #
    # TcEx Specific
    #

    batch_action: str = Field(
        'Create',
        description='The action for the Batch Job.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    batch_chunk: int = Field(
        25_000,
        description='The maximum number of item to send in batch request.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    batch_halt_on_error: bool = Field(
        False,
        description='Flag to control batch job failure behavior.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    batch_poll_interval: int = Field(
        15,
        description='The poll interval in second for the batch job.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    batch_poll_interval_max: int = Field(
        3_600,
        description='The maximum poll interval in seconds for the batch job.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    batch_write_type: str = Field(
        'Append',
        description='The API setting for batch to control write behavior.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )



"""TcEx Framework Module"""
# standard library
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v2.threat_intelligence.mapping.group.group import Group

if TYPE_CHECKING:
    # first-party
    from tcex.api.tc.v2.threat_intelligence.threat_intelligence import ThreatIntelligence


class Campaign(Group):
    """Unique API calls for Campaign API Endpoints

    Args:
        ti (ThreatIntelligence): An instance of the ThreatIntelligence Class.
        name (str, kwargs): [Required for Create] The name for this Group.
        owner (str, kwargs): The name for this Group. Default to default Org when not provided
        first_seen (str, kwargs): The first seen datetime expression for this Group.
    """

    def __init__(self, ti: 'ThreatIntelligence', **kwargs):
        """Initialize instance properties."""
        super().__init__(
            ti, sub_type='Campaign', api_entity='campaign', api_branch='campaigns', **kwargs
        )

    def first_seen(self, first_seen):
        """Update the campaign with the new first_seen date.

        Args:
            first_seen (str): The first_seen date. Converted to %Y-%m-%dT%H:%M:%SZ date format

        Returns:
            requests.Response: The response from the API call.
        """
        if not self.can_update():
            self._handle_error(910, [self.type])

        first_seen = self.util.any_to_datetime(first_seen).strftime('%Y-%m-%dT%H:%M:%SZ')
        self._data['firstSeen'] = first_seen
        request = {'firstSeen': first_seen}
        return self.tc_requests.update(self.api_type, self.api_branch, self.unique_id, request)



"""TcEx Framework Module"""
# standard library
from functools import cached_property

# third-party
import redis


class RedisClient:
    """A shared REDIS client connection using a Connection Pool.

    Initialize a single shared redis.connection.ConnectionPool.
    For a full list of kwargs see https://redis-py.readthedocs.io/en/latest/#redis.Connection.

    Args:
        host: The Redis host. Defaults to localhost.
        port: The Redis port. Defaults to 6379.
        db: The Redis db. Defaults to 0.
        blocking_pool: Use BlockingConnectionPool instead of ConnectionPool.
        errors (str, kwargs): The REDIS errors policy (e.g. strict).
        max_connections (int, kwargs): The maximum number of connections to REDIS.
        password (str, kwargs): The REDIS password.
        socket_timeout (int, kwargs): The REDIS socket timeout.
        timeout (int, kwargs): The REDIS Blocking Connection Pool timeout value.
    """

    def __init__(
        self,
        host: str = 'localhost',
        port: int = 6379,
        db: int = 0,
        blocking_pool: bool = False,
        **kwargs,
    ):
        """Initialize class properties"""
        password = kwargs.pop('password', None)
        username = kwargs.pop('username', None)

        pool = redis.ConnectionPool
        if blocking_pool:
            kwargs.pop('blocking_pool')  # remove blocking_pool key
            pool = redis.BlockingConnectionPool

        if username and password:
            self.pool = pool(
                host=host, port=port, db=db, username=username, password=password, **kwargs
            )
        else:
            self.pool = pool(host=host, port=port, db=db, **kwargs)

    @cached_property
    def client(self) -> redis.Redis:
        """Return an instance of redis.client.Redis."""
        return redis.Redis(connection_pool=self.pool)



"""TcEx Framework Module"""
# standard library
from collections.abc import Generator
from datetime import datetime

# third-party
import arrow as _arrow
from arrow import Arrow
from dateutil import parser
from dateutil.relativedelta import relativedelta


class DatetimeOperation:
    """TcEx Utilities Datetime Operations Class"""

    @classmethod
    def any_to_datetime(
        cls,
        datetime_expression: int | str | datetime | Arrow,
        tz: str | None = None,
    ) -> Arrow:
        """Return a arrow object from datetime expression.

        Args:
            datetime_expression: The datetime expression to parse into an Arrow datetime object.
            tz: If provided, the parsed Arrow datetime object will be converted to the timezone
                resulting from this timezone expression. Accepts values like 'US/Pacific', '-07:00',
                'UTC'.

                When this parameter is None, the returned Arrow datetime object will retain
                its timezone information if datetime_expression is timezone-aware. If
                datetime_expression is not timezone-aware, then the returned Arrow datetime object
                will have UTC timezone info.
        """
        value = str(datetime_expression)

        # note: order matters. For example, _parse_timestamp could parse inputs that would have
        # been meant for one of the default parser formats
        parser_methods = [
            cls._parse_default_arrow_formats,
            cls._parse_non_default_arrow_formats,
            cls._parse_timestamp,
            cls._parse_humanized_input,
            cls._parse_date_utils,
        ]

        for method in parser_methods:
            try:
                parsed = method(value)
            except Exception:  # nosec
                # value could not be parsed by current method.
                pass
            else:
                # convert timezone if tz arg provided, else return parsed object
                return cls._convert_timezone(parsed, tz) if tz is not None else parsed

        raise RuntimeError(  # pragma: no cover
            f'Value "{value}" of type "{type(datetime_expression)}" '
            'could not be parsed as a date time object.'
        )

    @property
    def arrow(self):
        """Return arrow for use in Apps."""
        return _arrow

    def chunk_date_range(
        self,
        start_date: int | str | datetime | Arrow,
        end_date: int | str | datetime | Arrow,
        chunk_size: int = 1,
        chunk_unit: str = 'months',
        date_format: str | None = None,
    ) -> Generator[tuple[Arrow | str, Arrow | str], None, None]:
        """Chunk a date range based on unit and size

        Args:
            start_date: Date time expression or datetime object.
            end_date: Date time expression or datetime object.
            chunk_size: Chunk size for the provided units.
            chunk_unit: A value of year, quarter, month, day, week, hour, minute, second (plural
            versions of valid values also acceptable)
            date_format: If None datetime object will be returned. Any other value
                must be a valid strftime format (%s for epoch seconds).
        """
        start_date = self.any_to_datetime(start_date)
        end_date = self.any_to_datetime(end_date)
        interval_args = {
            'frame': chunk_unit,
            'start': start_date,
            'end': end_date,
            'interval': chunk_size,
            'bounds': '[]',
            'exact': True,
        }

        try:
            for range_tuple in Arrow.interval(**interval_args):
                if date_format is not None:
                    range_tuple = range_tuple[0].strftime(date_format), range_tuple[1].strftime(
                        date_format
                    )
                yield range_tuple
        except Exception as ex:  # pragma: no cover
            raise RuntimeError(
                'Could not generate date range. Please verify that chunk_size, chunk_unit, '
                'and date_format values are valid.'
            ) from ex

    def timedelta(
        self, time_input1: int | str | datetime | Arrow, time_input2: int | str | datetime | Arrow
    ) -> dict:
        """Calculate the time delta between two time expressions.

        Args:
            time_input1: The end time expression (larger time value).
            time_input2: The start time input string (smaller time value).
        """
        time_input1_ = self.any_to_datetime(time_input1).datetime
        time_input2_ = self.any_to_datetime(time_input2).datetime

        diff = time_input1_ - time_input2_  # timedelta
        delta = relativedelta(time_input1_, time_input2_)  # relativedelta

        # totals
        total_months = (delta.years * 12) + delta.months
        total_weeks = (delta.years * 52) + (total_months * 4) + delta.weeks
        total_days = diff.days  # handles leap days
        total_hours = (total_days * 24) + delta.hours
        total_minutes = (total_hours * 60) + delta.minutes
        total_seconds = (total_minutes * 60) + delta.seconds
        total_microseconds = (total_seconds * 1000) + delta.microseconds
        return {
            'datetime_1': time_input1_.isoformat(),
            'datetime_2': time_input2_.isoformat(),
            'years': delta.years,
            'months': delta.months,
            'weeks': delta.weeks,
            'days': delta.days,
            'hours': delta.hours,
            'minutes': delta.minutes,
            'seconds': delta.seconds,
            'microseconds': delta.microseconds,
            'total_months': total_months,
            'total_weeks': total_weeks,
            'total_days': total_days,
            'total_hours': total_hours,
            'total_minutes': total_minutes,
            'total_seconds': total_seconds,
            'total_microseconds': total_microseconds,
        }

    @classmethod
    def _convert_timezone(cls, arrow_dt: Arrow, tz: str):
        """Convert Arrow datetime's timezone

        Args:
            arrow_dt: Arrow datetime object that will have its timezone converted
            tz: timezone expression. Accepts values like 'US/Pacific', '-07:00', 'UTC'.
        """
        try:
            return arrow_dt.to(tz)
        except Exception as ex:  # pragma: no cover
            raise RuntimeError(
                f'Could not convert datetime to timezone "{tz}". Please verify timezone input.'
            ) from ex

    @staticmethod
    def _parse_default_arrow_formats(value: str) -> Arrow:
        """Attempt to parse value using default Arrow formats.

        The value is simply passed into Arrow's "get" method. The following are the default
        date formats (found in arrow.parser.DateTimeParser.parse_iso):

        "YYYY-MM-DD",
        "YYYY-M-DD",
        "YYYY-M-D",
        "YYYY/MM/DD",
        "YYYY/M/DD",
        "YYYY/M/D",
        "YYYY.MM.DD",
        "YYYY.M.DD",
        "YYYY.M.D",
        "YYYYMMDD",
        # Year-Day-of-year: 2020-364 == 2020-12-29T00:00:00+00:00
        "YYYY-DDDD",
        # same as above, but without separating dash
        "YYYYDDDD",
        "YYYY-MM",
        "YYYY/MM",
        "YYYY.MM",
        "YYYY",
        # ISO week date: 2011-W05-4, 2019-W17
        "W"
        """
        return _arrow.get(value)

    @staticmethod
    def _parse_humanized_input(value: str) -> Arrow:
        """Attempt to dehumanize time inputs. Example: 'Two hours ago'."""
        now = _arrow.utcnow()
        plurals = {
            'second': 'seconds',
            'minute': 'minutes',
            'hour': 'hours',
            'day': 'days',
            'week': 'weeks',
            'month': 'months',
            'year': 'years',
        }

        value = value.lower().strip()
        if value == 'now':
            return now

        # pluralize singular time terms as applicable. Arrow does not support singular terms
        terms = [plurals.get(term, term) for term in value.split()]
        value = ' '.join(terms)

        return now.dehumanize(value)

    @staticmethod
    def _parse_non_default_arrow_formats(value: str) -> Arrow:
        """Attempt to parse value using non-default Arrow formats

        These are formats that Arrow provides constants for but are not used in the "get"
        method (Arrow method that parses values into datetimes) by default.

        Note: passing formats to test against overrides the default formats. Defaults are not used.
        """
        return _arrow.get(
            value,
            [
                _arrow.FORMAT_ATOM,
                _arrow.FORMAT_COOKIE,
                _arrow.FORMAT_RFC822,
                _arrow.FORMAT_RFC850,
                _arrow.FORMAT_RFC1036,
                _arrow.FORMAT_RFC1123,
                _arrow.FORMAT_RFC2822,
                _arrow.FORMAT_RFC3339,
                _arrow.FORMAT_RSS,
                _arrow.FORMAT_W3C,
            ],
        )

    @staticmethod
    def _parse_timestamp(value: str) -> Arrow:
        """Attempt to parse epoch timestamp in seconds, milliseconds, or microseconds.

        Note: passing formats to test against overrides the default formats. Defaults are not used.
        """
        # note: order matters. Must try to parse as milliseconds/microseconds first (x)
        # before trying to parse as seconds (X), else error occurs if passing a ms/ns value
        try:
            # attempt to parse as string first using microsecond/millisecond and second specifiers
            return _arrow.get(value, ['x', 'X'])
        except (_arrow.parser.ParserError, ValueError):
            # could not parse as string, try to parse as float
            return _arrow.get(float(value))

    @staticmethod
    def _parse_date_utils(value: str) -> Arrow:
        """Attempt to supplement arrows parsing ability with date util.

        Arrow doesn't support RFC_5322 used in HTTP 409 headers
        """
        return _arrow.get(parser.parse(value))



"""TcEx Framework Module"""
# standard library
import logging

from ...app.key_value_store import KeyValueRedis
from ...app.key_value_store.key_value_store import KeyValueStore

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class PlaybookDelete:
    """Playbook Delete"""

    def __init__(self, context: str, key_value_store: KeyValueStore):
        """Initialize the class properties."""
        self.context = context
        self.key_value_store = key_value_store

        # properties
        self.log = _logger

    def variable(self, key: str | None) -> int | None:
        """Delete method of CRUD operation for all data types.

        Only supported when using the Redis KV store.
        """
        if key is None:
            return None

        if not isinstance(self.key_value_store.client, KeyValueRedis):
            return None

        return self.key_value_store.client.delete(self.context, key.strip())



"""TcEx Framework Module"""
# standard library
import base64
import json
import logging
import re
from collections import OrderedDict
from typing import Any

from ...app.key_value_store.key_value_store import KeyValueStore
from ...registry import registry
from ...util.util import Util
from ...util.variable import BinaryVariable, StringVariable

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class PlaybookRead:
    """Playbook Read

    Args:
        key_value_store: A KV store instance.
        context: The KV Store context/session_id. For PB Apps the context is provided on
            startup, but for service Apps each request gets a different context.
        output_variables: The requested output variables. For PB Apps outputs are provided on
            startup, but for service Apps each request gets different outputs.
    """

    def __init__(self, context: str, key_value_store: KeyValueStore):
        """Initialize the class properties."""
        self.context = context
        self.key_value_store = key_value_store

        # properties
        self.log = _logger
        self.util = Util()

    def _check_variable_type(self, variable: str, type_: str):
        """Validate the correct type was passed to the method."""
        if self.util.get_playbook_variable_type(variable).lower() != type_.lower():
            raise RuntimeError(
                f'Invalid variable provided ({variable}), variable must be of type {type_}.'
            )

    @staticmethod
    def _coerce_string_value(value: bool | float | int | str) -> str:
        """Return a string value from an bool or int."""
        # coerce bool before int as python says a bool is an int
        if isinstance(value, bool):
            # coerce bool to str type
            value = str(value).lower()

        # coerce int to str type
        if isinstance(value, float | int):
            value = str(value)

        return value

    @staticmethod
    def _decode_binary(data: bytes) -> str:
        """Return decoded bytes data handling data written by java apps."""
        try:
            _data = data.decode('utf-8')
        except UnicodeDecodeError:  # pragma: no cover
            # for data written an upstream java App
            _data = data.decode('latin-1')
        return _data

    @staticmethod
    def _deserialize_data(value: bytes | str) -> Any:
        """Return the loaded JSON value or raise an error."""
        try:
            return json.loads(value, object_pairs_hook=OrderedDict)
        except ValueError as ex:  # pragma: no cover
            raise RuntimeError(f'Failed to JSON load data "{value}" ({ex}).') from ex

    def _get_data(self, key: str) -> bytes | str | None:
        """Get the value from Redis if applicable."""
        try:
            return self.key_value_store.client.read(self.context, key.strip())
        except RuntimeError as ex:
            self.log.error(ex)
        return None

    @staticmethod
    def _load_data(value: str) -> dict | list[dict | str] | str:
        """Return the loaded JSON value or raise an error."""
        try:
            return json.loads(value, object_pairs_hook=OrderedDict)
        except ValueError as ex:  # pragma: no cover
            raise RuntimeError(f'Failed to JSON load data "{value}" ({ex}).') from ex

    def _null_key_check(self, key: Any) -> bool:
        """Return False if value is not null."""
        if key is None:
            self.log.warning('The provided key was None.')
            return True

        return False

    def _process_key_value(self, data: dict, resolve_embedded: bool) -> dict | None:
        """Read the value from key value store.

        KeyValue data should be stored as a JSON string.
        """
        # IMPORTANT:
        # A Single level of nested variables is supported. There is no way
        # in the TC platform to create double nested variables. Any App
        # that would try and create a double nested variable is improperly
        # written.

        # KeyValue List Input
        # -------------------------------------------------------------
        # | key                         | value                       |
        # =============================================================
        # | my_binary                   | #App:7979:two!Binary        |
        # -------------------------------------------------------------
        # | my_binary_array             | #App:7979:two!BinaryArray   |
        # -------------------------------------------------------------
        # | my_key_value                | #App:7979:two!KeyValue      |
        # -------------------------------------------------------------
        # | my_key_value_array          | #App:7979:two!KeyValueArray |
        # -------------------------------------------------------------
        # | my_string                   | #App:7979:two!String        |
        # -------------------------------------------------------------
        # | my_string_array             | #App:7979:two!StringArray   |
        # -------------------------------------------------------------
        # | my_tcentity                 | #App:7979:two!TCEntity      |
        # -------------------------------------------------------------
        # | my_tcentity_array           | #App:7979:two!TCEntityArray |
        # -------------------------------------------------------------

        # An upstream Apps KeyValue output can be used in a KeyValueList input, but
        # Apps SHOULD NOT be writing KeyValueArray with nested variables. This means
        # that there will only ever be 1 levels of nesting.

        # KeyValueList Input -> Nested KeyValue/KeyValueArray, the nested
        # KeyValue/KeyValueArray CAN NOT have nested variables since they
        # have to come from an upstream App.

        # check if keyvalue value is a variable
        if resolve_embedded:
            value = data['value']
            if self.util.is_playbook_variable(value):
                # any type can be nested, but no further nesting is supported
                data['value'] = self.any(value)
            else:
                # read embedded is less efficient and has more caveats
                data['value'] = self._read_embedded(value)

        return data

    @staticmethod
    def _process_space_patterns(string: str) -> str:
        r"""Return the string with \s replace with spaces."""
        # replace "\s" with a space only for user input.
        # using '\\s' will prevent replacement.
        string = re.sub(r'(?<!\\)\\s', ' ', string)
        string = re.sub(r'\\\\s', r'\\s', string)
        return string

    def _read_embedded(self, value: str) -> str:
        r"""Read method for "embedded" variables.

        .. Note:: The ``read()`` method will automatically determine if the input is a variable or
            needs to be searched for embedded variables.

        Embedded variable rules:

        * Only user input can have embedded variables.
        * Only String and KeyValueArray variables can have embedded variables.
        * Variables can only be embedded one level deep.

        This method will automatically convert variables embedded in a string with value retrieved
        from DB. If there are no keys/variables the raw string will be returned.

        Examples::

            DB Values
            #App:7979:variable_name!String:
                "embedded \\"variable\\""
            #App:7979:two!String:
                "two"
            #App:7979:variable_name!StringArray:
                ["one", "two", "three"]

            Examples 1:
                Input:  "This input has a embedded #App:7979:variable_name!String"

            Examples 2:
                Input: ["one", #App:7979:two!String, "three"]

            Examples 3:
                Input: [{
                    "key": "embedded string",
                    "value": "This input has a embedded #App:7979:variable_name!String"
                }, {
                    "key": "string array",
                    "value": #App:7979:variable_name!StringArray
                }, {
                    "key": "string",
                    "value": #App:7979:variable_name!String
                }]

        Args:
            value (str): The value to parsed and updated from the DB.

        Returns:
            (str): Results retrieved from DB
        """
        if value is None:  # pragma: no cover
            return value

        for match in re.finditer(self.util.variable_expansion_pattern, str(value)):
            variable = match.group(0)  # the full variable pattern
            v = None
            if match.group('origin') == '#':  # pb-variable
                v = self.any(variable)
            elif match.group('origin') == '&':  # tc-variable
                v = registry.inputs.resolve_variable(variable)

            # TODO: [high] should this behavior be changed in 3.0?
            self.log.debug(f'embedded variable: {variable}, value: {v}')

            if match.group('type') in ['Binary', 'BinaryArray']:
                self.log.debug(
                    f'Binary types may not be embedded into strings. Could not embed: {variable}'
                )
                v = '<binary>'

            if isinstance(v, dict | list):
                v = json.dumps(v)
            elif v is None:
                v = '<null>'

            # value.replace was chosen over re.sub due to an issue encountered while testing an app.
            # re.sub will handle escaped characters like \t, value.replace would not handle these
            # scenarios.
            if isinstance(v, str):
                value = value.replace(variable, v)

        return StringVariable(value)

    @staticmethod
    def _to_array(value: list | str | None) -> list:
        """Return the provided array as a list."""
        if value is None:
            # Adding none value to list breaks App logic. It's better to not request
            # Array and build array externally if None values are required.
            value = []
        elif not isinstance(value, list):
            value = [value]
        return value

    def any(self, key: str) -> bytes | dict | list | str | None:
        """Return the value from the keystore for all types.

        This is a quick helper method, for more advanced features
        the individual read methods should be used (e.g., binary).
        """
        if self._null_key_check(key) is True:
            return None

        key = key.strip()  # clean up key
        variable_type = self.util.get_playbook_variable_type(key).lower()
        variable_type_map = {
            'binary': self.binary,
            'binaryarray': self.binary_array,
            'keyvalue': self.key_value,
            'keyvaluearray': self.key_value_array,
            'string': self.string,
            'stringarray': self.string_array,
            'tcbatch': self.tc_batch,
            'tcentity': self.tc_entity,
            'tcentityarray': self.tc_entity_array,
            # 'tcenhancedentity': self.tc_entity,
        }
        value = variable_type_map.get(variable_type, self.raw)(key)

        if value is not None:
            if variable_type == 'binary':
                value = BinaryVariable(value)
            elif variable_type == 'binaryarray':
                value = [v if v is None else BinaryVariable(v) for v in value]
            elif variable_type == 'string':
                value = StringVariable(value)
            elif variable_type == 'stringarray':
                value = [v if v is None else StringVariable(v) for v in value]

        return value

    def binary(
        self,
        key: str,
        b64decode: bool = True,
        decode: bool = False,
    ) -> BinaryVariable | str | None:
        """Read the value from key value store.

        The binary write method base64 encodes the data, then decodes the bytes to string, and
        finally serializes the string before writing to the key value store.

        This method will deserialize the string, then OPTIONALLY base64 decode the data, and
        finally return the Binary data.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'Binary')

        # get the data from the key value store
        data = self._get_data(key)
        if data is None:
            return None

        # reverse the order of the binary create/write method
        # 1. deserialize the data
        # 2. base64 decode the data

        # deserialize the data
        data = self._deserialize_data(data)

        # base64 decode the data (get_data returns multiple types, but the binary
        # write method will always write a base64.encoded->bytes.decoded->serialized string)
        # for the testing framework, the base64 encoded string should be returned so that
        # the data can be compared to the expected value stored in the test profile.
        if b64decode is True and isinstance(data, str):
            data = BinaryVariable(base64.b64decode(data))
            if decode is True:
                # allow developer to decided if they want bytes or str
                data = self._decode_binary(data)
        elif isinstance(data, bytes):
            # data should never be returned as bytes, but just in case an old App is using an
            # older version of TcEx or if the TC Platform is writes binary data to the key value
            # store, decode the bytes to a string
            data = self._decode_binary(data)

        return data

    def binary_array(
        self,
        key: str,
        b64decode: bool = True,
        decode: bool = False,
    ) -> list[BinaryVariable | str] | None:
        """Read the value from key value store.

        The binary array write method iterates over the BinaryArray and base64 encodes the data,
        then decodes the bytes to string, and finally serializes the array before writing to the
        key value store.

        This method will deserialize the string, then iterate over the array and OPTIONALLY base64
        decode the data, and finally return the BinaryArray.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'BinaryArray')

        # get the data from the key value store
        data = self._get_data(key)
        if data is None:
            return None

        # reverse the order of the binary create/write method
        # 1. deserialize the data
        # 2. iterate over the array
        # 3. base64 decode the data

        # data should be a serialized string, but in case there are any legacy Apps that are
        # using an older version of TcEx, check for bytes and decode to string
        if isinstance(data, bytes):
            data = data.decode('utf-8')

        # deserialize the data
        _data: list[str] = self._deserialize_data(data)

        values = []
        for d in _data:
            if b64decode is True and isinstance(d, str):
                d = BinaryVariable(base64.b64decode(d))
                if decode is True:
                    # allow developer to decided if they want bytes or str
                    d = self._decode_binary(d)
            values.append(d)
        return values

    def key_value(
        self,
        key: str,
        resolve_embedded: bool = True,
    ) -> dict | None:
        """Read the value from key value store.

        KeyValue data should be stored as a JSON string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'KeyValue')

        # get the data from the key value store
        data = self._get_data(key)
        if data is None:
            return None

        # deserialize the data
        data = self._deserialize_data(data)

        return self._process_key_value(data, resolve_embedded=resolve_embedded)

    def key_value_array(
        self,
        key: str,
        resolve_embedded: bool = True,
    ) -> list[dict] | None:
        """Read the value from key value store.

        KeyValueArray data should be stored as serialized string.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'KeyValueArray')

        data = self._get_data(key)
        if data is None:
            return None

        # data should be a serialized string, but in case there are any legacy Apps that are
        # using an older version of TcEx, check for bytes and decode to string
        if isinstance(data, bytes):
            data = data.decode()

        # Array type is serialized before writing to redis, deserialize the data
        _data: list[dict] = self._deserialize_data(data)

        values = []
        for d in _data:
            # d should be a base64 encoded string
            values.append(self._process_key_value(d, resolve_embedded=resolve_embedded))
        return values

    def raw(self, key: str) -> Any | None:
        """Read method of CRUD operation for raw data.

        Bytes input will be returned a as string as there is no way
        to determine data from redis originated as bytes or string.
        """
        if self._null_key_check(key) is True:
            return None

        return self.key_value_store.client.read(self.context, key.strip())

    def string(
        self,
        key: str,
        resolve_embedded: bool = True,
    ) -> StringVariable | None:
        """Read the value from key value store.

        The string write method serializes the string before writing to the key value store.

        This method will deserialize the string and finally return the StringArray data.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'String')

        # get the data from the key value store
        data = self._get_data(key)
        if data is None:
            return None

        # data should be a serialized string, but in case there are any legacy Apps that are
        # using an older version of TcEx, check for bytes and decode to string
        if isinstance(data, bytes):
            data = data.decode()

        # deserialize the data
        data = self._deserialize_data(data)

        # only resolve embedded variables if resolve_embedded is True and
        # the entire string does not exactly match a variable pattern
        if resolve_embedded and not self.util.is_playbook_variable(data):
            data = self._read_embedded(data)

        # coerce data back to string, since technically TC doesn't support bool, int, etc
        return StringVariable(self._coerce_string_value(data))

    def string_array(self, key: str) -> list[StringVariable] | None:
        """Read the value from key value store.

        The string_array write method serializes the list of strings before writing to the key value
        store.

        This method will deserialize the list of strings and finally return the StringArray data.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'StringArray')

        # get the data from the key value store
        data = self._get_data(key)
        if data is None:
            return None

        # data should be a serialized string, but in case there are any legacy Apps that are
        # using an older version of TcEx, check for bytes and decode to string
        if isinstance(data, bytes):
            data = data.decode()

        # deserialize the data
        _data: list[str] = self._deserialize_data(data)

        # return array of StringVariables
        # return [StringVariable(self._coerce_string_value(d)) for d in _data]
        return [d if d is None else StringVariable(self._coerce_string_value(d)) for d in _data]

    def tc_batch(self, key: str) -> dict | None:
        """Read the value from key value store.

        The tc_batch write method serializes the string before writing to the key value store.

        This method will deserialize the string and finally return the TCBatch data.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'TCBatch')

        data = self._get_data(key)
        if data is None:
            return None

        # data should be a serialized string, but in case there are any legacy Apps that are
        # using an older version of TcEx, check for bytes and decode to string
        if isinstance(data, bytes):
            data = data.decode()

        return self._deserialize_data(data)

    def tc_entity(self, key: str) -> dict[str, str] | None:
        """Read the value from key value store.

        The tc_entity write method serializes the dict before writing to the key value store.

        This method will deserialize the string and finally return the TCEntity data.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'TCEntity')

        # get the data from the key value store
        data = self._get_data(key)
        if data is None:
            return None

        # data should be a serialized string, but in case there are any legacy Apps that are
        # using an older version of TcEx, check for bytes and decode to string
        if isinstance(data, bytes):
            data = data.decode()

        # deserialize the data
        return self._deserialize_data(data)

    def tc_entity_array(
        self,
        key: str,
    ) -> list[dict[str, str]] | None:
        """Read the value from key value store.

        The tc_entity_array write method serializes the list of dicts before writing to the key
        value store.

        This method will deserialize the list of dicts and finally return the TCEntityArray data.
        """
        if self._null_key_check(key) is True:
            return None

        # quick check to ensure an invalid key was not provided
        self._check_variable_type(key, 'TCEntityArray')

        # get the data from the key value store
        data = self._get_data(key)
        if data is None:
            return None

        # data should be a serialized string, but in case there are any legacy Apps that are
        # using an older version of TcEx, check for bytes and decode to string
        if isinstance(data, bytes):
            data = data.decode()

        # deserialize the data
        return self._deserialize_data(data)

    def variable(self, key: str | None, array: bool = False) -> bytes | dict | list | str | None:
        """Read method of CRUD operation for working with KeyValue DB.

        This method will automatically check to see if a single variable is passed
        or if "mixed" data is passed and return the results from the DB. It will also
        automatically determine the variable type to read.
        """
        value = key
        if value is not None and isinstance(key, str):
            key = key.strip()

            if re.match(self.util.variable_playbook_match, key):
                value = self.any(key=key)
            else:
                # replace space patterns
                value = self._process_space_patterns(value)

                # key must be an embedded variable
                value = self._read_embedded(value)

        if array is True:
            if isinstance(value, list | str):
                value = self._to_array(value)
            elif value is None:
                value = []

        return value



"""TcEx Framework Module"""
# third-party
from pydantic import BaseModel, Field

# first-party
from tcex.input.field_type.sensitive import Sensitive


class ServiceModel(BaseModel):
    """Input Service Model

    Supported runtimeLevel:
    * ApiService
    * TriggerService
    * WebhookTriggerService
    """

    tc_svc_broker_cacert_file: str = Field(
        None,
        description='The Broker SSL CA (full chain) certificate.',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_broker_cert_file: str = Field(
        None,
        description='The Broker SSL Server certificate.',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_broker_conn_timeout: int = Field(
        60,
        description='The broker connection startup timeout in seconds.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_svc_broker_host: str = Field(
        None,
        description='The Broker service hostname.',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_broker_jks_file: str | None = Field(
        'Unused',
        description='Input for Java Apps.',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_broker_jks_pwd: str | None = Field(
        'Unused',
        description='Input for Java Apps.',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_broker_port: int = Field(
        None,
        description='The Broker service port number.',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_broker_timeout: int = Field(
        60,
        description='The broker service timeout in seconds.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_svc_broker_token: Sensitive = Field(
        None,
        description='The Broker auth token.',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_client_topic: str = Field(
        None,
        description='The Broker client topic (App -> Core).',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_hb_timeout_seconds: int = Field(
        20,
        description='The heartbeat timeout interval in seconds.',
        inclusion_reason='runtimeLevel',
    )
    tc_svc_id: int = Field(
        None,
        description='The unique ID for the current service.',
        inclusion_reason='runtimeLevel',
        requires_definition=True,
    )
    tc_svc_server_topic: str = Field(
        None,
        description='The Broker server topic (Core -> App).',
        inclusion_reason='runtimeLevel',
    )
    tcex_testing_context: str | None = Field(
        None,
        description='[Testing] The testing framework context.',
        inclusion_reason='runtimeLevel',
    )



"""TcEx Framework Module"""
# standard library
import json
import logging
from collections import OrderedDict
from functools import cached_property
from pathlib import Path

from ...pleb.singleton import Singleton
from .model.job_json_model import JobJsonModel

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class JobJson(metaclass=Singleton):
    """Config object for job.json file"""

    def __init__(
        self,
        filename: str | None = None,
        path: Path | str | None = None,
        logger: logging.Logger | None = None,
    ):
        """Initialize instance properties."""
        filename = filename or 'tcex.json'
        path = Path(path or Path.cwd())
        self.log = logger or _logger

        # properties
        self.fqfn = path / filename

    @cached_property
    def contents(self) -> dict:
        """Return job.json file contents."""
        _contents = {}

        if self.fqfn.is_file():
            try:
                with self.fqfn.open() as fh:
                    _contents = json.load(fh, object_pairs_hook=OrderedDict)
            except OSError:  # pragma: no cover
                self.log.error(
                    f'feature=tcex-json, exception=failed-reading-file, filename={self.fqfn}'
                )
        else:  # pragma: no cover
            self.log.error(f'feature=tcex-json, exception=file-not-found, filename={self.fqfn}')

        return _contents

    @cached_property
    def model(self) -> JobJsonModel:
        """Return the Install JSON model."""
        return JobJsonModel(**self.contents)



"""TcEx Framework Module"""
# standard library
from collections.abc import Generator, Iterator
from typing import TYPE_CHECKING

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.object_abc import ObjectABC
from tcex.api.tc.v3.object_collection_abc import ObjectCollectionABC
from tcex.api.tc.v3.security_labels.security_label_model import SecurityLabelModel
from tcex.api.tc.v3.tags.tag_model import TagModel
from tcex.api.tc.v3.victim_assets.victim_asset_model import VictimAssetModel
from tcex.api.tc.v3.victim_attributes.victim_attribute_model import VictimAttributeModel
from tcex.api.tc.v3.victims.victim_filter import VictimFilter
from tcex.api.tc.v3.victims.victim_model import VictimModel, VictimsModel

if TYPE_CHECKING:  # pragma: no cover
    # first-party
    from tcex.api.tc.v3.groups.group import Group  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.security_labels.security_label import SecurityLabel  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.tags.tag import Tag  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.victim_assets.victim_asset import VictimAsset  # CIRCULAR-IMPORT
    from tcex.api.tc.v3.victim_attributes.victim_attribute import VictimAttribute  # CIRCULAR-IMPORT


class Victim(ObjectABC):
    """Victims Object.

    Args:
        assets (VictimAssets, kwargs): A list of victim assets corresponding to the Victim.
        associated_groups (Groups, kwargs): A list of groups that this victim is associated with.
        attributes (VictimAttributes, kwargs): A list of Attributes corresponding to the Victim.
        name (str, kwargs): Name of the Victim.
        nationality (str, kwargs): Nationality of the Victim.
        org (str, kwargs): Org of the Victim.
        security_labels (SecurityLabels, kwargs): A list of Security Labels corresponding to the
            Intel item (NOTE: Setting this parameter will replace any existing tag(s) with
            the one(s) specified).
        suborg (str, kwargs): Suborg of the Victim.
        tags (Tags, kwargs): A list of Tags corresponding to the item (NOTE: Setting this parameter
            will replace any existing tag(s) with the one(s) specified).
        work_location (str, kwargs): Work location of the Victim.
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(kwargs.pop('session', None))

        # properties
        self._model: VictimModel = VictimModel(**kwargs)
        self._nested_field_name = 'victims'
        self._nested_filter = 'has_victim'
        self.type_ = 'Victim'

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.VICTIMS.value

    @property
    def model(self) -> VictimModel:
        """Return the model data."""
        return self._model

    @model.setter
    def model(self, data: dict | VictimModel):
        """Create model using the provided data."""
        if isinstance(data, type(self.model)):
            # provided data is already a model, nothing required to change
            self._model = data
        elif isinstance(data, dict):
            # provided data is raw response, load the model
            self._model = type(self.model)(**data)
        else:
            raise RuntimeError(f'Invalid data type: {type(data)} provided.')

    @property
    def as_entity(self) -> dict:
        """Return the entity representation of the object."""
        type_ = self.type_

        return {'type': type_, 'id': self.model.id, 'value': self.model.name}

    @property
    def victim_assets(self) -> Generator['VictimAsset', None, None]:
        """Yield VictimAsset from VictimAssets."""
        # first-party
        from tcex.api.tc.v3.victim_assets.victim_asset import VictimAssets

        yield from self._iterate_over_sublist(VictimAssets)  # type: ignore

    @property
    def associated_groups(self) -> Generator['Group', None, None]:
        """Yield Group from Groups."""
        # first-party
        from tcex.api.tc.v3.groups.group import Groups

        yield from self._iterate_over_sublist(Groups)  # type: ignore

    @property
    def attributes(self) -> Generator['VictimAttribute', None, None]:
        """Yield Attribute from Attributes."""
        # first-party
        from tcex.api.tc.v3.victim_attributes.victim_attribute import VictimAttributes

        yield from self._iterate_over_sublist(VictimAttributes)  # type: ignore

    @property
    def security_labels(self) -> Generator['SecurityLabel', None, None]:
        """Yield SecurityLabel from SecurityLabels."""
        # first-party
        from tcex.api.tc.v3.security_labels.security_label import SecurityLabels

        yield from self._iterate_over_sublist(SecurityLabels)  # type: ignore

    @property
    def tags(self) -> Generator['Tag', None, None]:
        """Yield Tag from Tags."""
        # first-party
        from tcex.api.tc.v3.tags.tag import Tags

        yield from self._iterate_over_sublist(Tags)  # type: ignore

    def stage_victim_asset(self, data: dict | ObjectABC | VictimAssetModel):
        """Stage victim_asset on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = VictimAssetModel(**data)

        if not isinstance(data, VictimAssetModel):
            raise RuntimeError('Invalid type passed in to stage_victim_asset')
        data._staged = True
        self.model.assets.data.append(data)  # type: ignore

    def stage_attribute(self, data: dict | ObjectABC | VictimAttributeModel):
        """Stage attribute on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = VictimAttributeModel(**data)

        if not isinstance(data, VictimAttributeModel):
            raise RuntimeError('Invalid type passed in to stage_attribute')
        data._staged = True
        self.model.attributes.data.append(data)  # type: ignore

    def stage_security_label(self, data: dict | ObjectABC | SecurityLabelModel):
        """Stage security_label on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = SecurityLabelModel(**data)

        if not isinstance(data, SecurityLabelModel):
            raise RuntimeError('Invalid type passed in to stage_security_label')
        data._staged = True
        self.model.security_labels.data.append(data)  # type: ignore

    def stage_tag(self, data: dict | ObjectABC | TagModel):
        """Stage tag on the object."""
        if isinstance(data, ObjectABC):
            data = data.model  # type: ignore
        elif isinstance(data, dict):
            data = TagModel(**data)

        if not isinstance(data, TagModel):
            raise RuntimeError('Invalid type passed in to stage_tag')
        data._staged = True
        self.model.tags.data.append(data)  # type: ignore


class Victims(ObjectCollectionABC):
    """Victims Collection.

    # Example of params input
    {
        'result_limit': 100,  # Limit the retrieved results.
        'result_start': 10,  # Starting count used for pagination.
        'fields': ['caseId', 'summary']  # Select additional return fields.
    }

    Args:
        session (Session): Session object configured with TC API Auth.
        tql_filters (list): List of TQL filters.
        params (dict): Additional query params (see example above).
    """

    def __init__(self, **kwargs):
        """Initialize instance properties."""
        super().__init__(
            kwargs.pop('session', None), kwargs.pop('tql_filter', None), kwargs.pop('params', None)
        )
        self._model = VictimsModel(**kwargs)
        self.type_ = 'victims'

    def __iter__(self) -> Iterator[Victim]:
        """Return CM objects."""
        return self.iterate(base_class=Victim)  # type: ignore

    @property
    def _api_endpoint(self) -> str:
        """Return the type specific API endpoint."""
        return ApiEndpoints.VICTIMS.value

    @property
    def filter(self) -> VictimFilter:
        """Return the type specific filter object."""
        return VictimFilter(self.tql)



"""TcEx Framework Module"""
# standard library
import json
import logging

# third-party
from requests import Session

# first-party
from tcex.exit.error_code import handle_error
from tcex.logger.trace_logger import TraceLogger

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class Notification:
    """TcEx Notification Class"""

    def __init__(self, session_tc: Session):
        """Initialize the Class properties.

        Args:
            session_tc: An configured instance of request.Session with TC API Auth.
        """
        self.session_tc = session_tc

        # properties
        self._is_organization = False
        self._notification_type = None
        self._recipients = None
        self._priority = 'Low'
        self.log = _logger

    def recipients(self, notification_type: str, recipients: str, priority='Low'):
        """Set vars for the passed in data. Used for one or more recipient notification.

        .. code-block:: javascript

            {
                "notificationType": notification_type,
                "priority": priority
                "isOrganization": false,
                "recipients": recipients
            }

        Args:
            notification_type: The type of notification being sent.
            recipients: A comma delimited string of recipients.
            priority: The priority: Low, Medium, High.
        """
        self._notification_type = notification_type
        self._recipients = recipients
        self._priority = priority
        self._is_organization = False

    def org(self, notification_type: str, priority: str = 'Low'):
        """Set vars for the passed in data. Used for org notification.

        .. code-block:: javascript

            {
                "notificationType": notification_type,
                "priority": priority
                "isOrganization": true
            }

        Args:
            notification_type: The notification type.
            priority: The priority: Low, Medium, High.
        """
        self._notification_type = notification_type
        self._recipients = None
        self._priority = priority
        self._is_organization = True

    def send(self, message: str):
        """Send our message

        Args:
            message: The message to be sent.

        Returns:
            requests.models.Response: The response from the request.

        """
        body = {
            'notificationType': self._notification_type,
            'priority': self._priority,
            'isOrganization': self._is_organization,
            'message': message,
        }

        if self._recipients:
            body['recipients'] = self._recipients

        self.log.debug(f'notification body: {json.dumps(body)}')

        # create our tcex resource
        r = self.session_tc.post('/v2/notifications', json=body)
        if r.status_code == 400:
            # specifically handle unknown users
            self.log.error(f'Failed to send notification ({r.text})')
        elif not r.ok:  # pragma: no cover
            handle_error(750, [r.status_code, r.text])

        # return response body
        return r.json()



"""TcEx Framework Module"""

# first-party
from tcex.input.model.common_model import CommonModel
from tcex.input.model.playbook_common_model import PlaybookCommonModel
from tcex.input.model.playbook_model import PlaybookModel
from tcex.input.model.service_model import ServiceModel


# pylint: disable=too-many-ancestors
class CommonAdvancedModel(CommonModel, PlaybookCommonModel, PlaybookModel, ServiceModel):
    """Model Definition"""



"""TcEx Framework Module"""
# standard library
import json
import logging
import os
import random
import re
import sys
from collections.abc import Iterator
from pathlib import Path
from typing import Any, cast

try:
    # standard library
    import sqlite3
except ImportError:  # pragma: no cover
    pass  # sqlite3 is only required for local development

from ...pleb.cached_property import cached_property
from ...pleb.none_model import NoneModel
from ...util.render.render import Render
from .install_json import InstallJson
from .layout_json import LayoutJson
from .model.install_json_model import OutputVariablesModel, ParamsModel
from .model.layout_json_model import OutputsModel

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class InputModel(ParamsModel):
    """Input Model"""

    value: Any

    def __hash__(self):
        """Make model hashable."""
        return hash(self.name)


class Permutation:
    """Permutations Module

    Calculate permutations for inputs and outputs based on layout.json and install.json files.
    """

    def __init__(self):
        """Initialize instance properties"""

        # properties
        self._input_names = None
        self._input_permutations: list[list[InputModel]] = []
        self._output_permutations: list[list[OutputVariablesModel]] = []
        self.fqfn = Path(os.getcwd(), 'permutations.json')
        self.ij = InstallJson()
        self.input_table = 'inputs'
        self.lj = LayoutJson()
        self.log = _logger

    @staticmethod
    def _create_input_model(ij_param: ParamsModel, value: Any) -> InputModel:
        """Create an input model from the install.json param model."""
        _input_model = InputModel(**ij_param.dict())
        # manually adding List values due to bug where data is not getting loaded into model in init
        _input_model.intel_type = ij_param.intel_type
        _input_model.playbook_data_type = ij_param.playbook_data_type
        _input_model.valid_values = ij_param.valid_values
        _input_model.value = value
        return _input_model

    @cached_property
    def _display_keywords(self) -> set[str]:
        """Return the display keywords."""
        _keywords = set()
        for param in self.lj.model.params.values():
            for keyword in (param.display or '').split(' '):
                keyword = re.sub(r'[^a-zA-Z0-9_]', '', keyword)
                if keyword in self.ij.model.param_names:
                    _keywords.add(keyword)
        self.log.debug(f'keywords={_keywords}')
        return _keywords

    def _gen_permutations(self, index: int = 0, params: list | None = None):
        """Iterate recursively over layout.json parameter names to build permutations.

        .. NOTE:: Permutations are for layout.json based Apps.

        Args:
            index: The current index position in the layout names list.
            params: The current list of args.
        """
        params = params or []
        try:
            # grab the name using the index value of all params in the layout.json file.
            # after the last name is hit the IndexError will trigger collecting outputs.
            name = list(self.lj.model.param_names)[index]

            # get layout.json param name and data
            lj_param = self.lj.model.get_param(name)

            # get install.json param to match layout.json param
            ij_param = self.ij.model.get_param(name)
            if ij_param is None:  # pragma: no cover
                Render.panel.failure(f'No param found in install.json for "{name}".')

            if self.validate_layout_display(self.input_table, lj_param.display) or ij_param.hidden:
                # only process params that match display query or are hidden
                if name != 'tc_action' and name not in self._display_keywords:
                    params.append(self._create_input_model(ij_param, None))

                    # recursively call method to get all permutations
                    self._gen_permutations(index + 1, list(params))
                elif ij_param.type.lower() == 'boolean':
                    for val in [True, False]:
                        params.append(self._create_input_model(ij_param, val))

                        # update the data in the sqlite db so for next iteration
                        self.db_update_record(self.input_table, name, val)

                        # recursively call method to get all permutations
                        self._gen_permutations(index + 1, list(params))

                        # remove the previous arg before next iteration
                        params.pop()
                elif ij_param.type.lower() in ['choice', 'editchoice']:
                    for val in self.ij.expand_valid_values(ij_param.valid_values):
                        params.append(self._create_input_model(ij_param, val))

                        # update the data in the sqlite db so for next iteration
                        self.db_update_record(self.input_table, name, val)

                        # recursively call method to get all permutations
                        self._gen_permutations(index + 1, list(params))

                        # remove the previous arg before next iteration
                        params.pop()
                else:
                    params.append(self._create_input_model(ij_param, None))

                    # recursively call method to get all permutations
                    self._gen_permutations(index + 1, list(params))
            else:
                # do not add param since it's not required for this permutation
                self._gen_permutations(index + 1, list(params))
        except IndexError:
            # when IndexError is reached all params has been processed
            self._input_permutations.append(params)
            outputs: list[OutputVariablesModel] = []

            # iterate of InstallJsonModel -> PlaybookModel -> OutputVariablesModel
            if self.ij.model.playbook is not None:
                for o in self.ij.model.playbook.output_variables:
                    # get layout.json param to match install.json output variable
                    lj_output: NoneModel | OutputsModel = self.lj.model.get_output(o.name)
                    if isinstance(lj_output, OutputsModel):
                        valid = self.validate_layout_display(self.input_table, lj_output.display)
                        if lj_output.display is None or not valid:
                            continue
                    # output meet permutation check
                    outputs.append(o)
                self._output_permutations.append(outputs)

    @property
    def _params_data(self) -> Iterator[ParamsModel]:
        """Return all defined params from layout.json/install.json, including hidden params."""
        # using inputs from layout.json since they are required to be in order
        # (display field can only use inputs previously defined)
        for input_name in self.lj.model.params:
            # get data from install.json based on name
            param = self.ij.model.get_param(input_name)
            if param is not None:
                yield param

        # hidden fields will not be in layout.json so they need to be include manually
        for input_name, ij_data in self.ij.model.filter_params(hidden=True).items():
            yield ij_data

    @cached_property
    def action_configurations(self) -> dict:
        """Return action configuration."""
        self.init_permutations()

        _action_configurations = {}
        for index, inputs in enumerate(self._input_permutations):
            for input_ in inputs:
                if input_.name == 'tc_action':
                    # the value should always be a string
                    action = cast(str, input_.value)
                    _action_configurations.setdefault(action, {'inputs': [], 'outputs': []})
                    _action_configurations[action]['inputs'].extend(inputs)
                    _action_configurations[action]['outputs'].extend(
                        self._output_permutations[index]
                    )
                    break

        for action, data in _action_configurations.items():
            _action_configurations[action] = {
                'inputs': sorted(list(set(data['inputs'])), key=lambda x: x.name),
                'outputs': sorted(list(set(data['outputs'])), key=lambda x: x.name),
            }

        return _action_configurations

    @cached_property
    def db_conn(self) -> sqlite3.Connection:  # type: ignore
        """Create a temporary in memory DB and return the connection."""
        try:
            return sqlite3.connect(':memory:')
        except sqlite3.Error as ex:  # pragma: no cover
            Render.panel.failure(f'Failed initializing database ({ex}).')

    def db_create_table(self, table_name: str, columns: list[str]):
        """Create a temporary DB table.

        Args:
            table_name: The DB table name.
            columns: The DB column names.
        """
        columns_ = ', '.join([f'''"{c.strip('"').strip("'")}" text''' for c in set(columns)])
        self.log.debug(f'action=db-create-table, table-name={table_name}, columns={columns_}')
        sql = f'CREATE TABLE IF NOT EXISTS {table_name} ({columns_});'
        try:
            cr = self.db_conn.cursor()
            cr.execute(sql)
        except sqlite3.Error as e:  # pragma: no cover
            Render.panel.failure(f'SQL create db failed - SQL: "{sql}", Error: "{e}"')

    def db_drop_table(self, table_name: str):
        """Drop a DB table.

        Args:
            table_name: The DB table name.
        """
        sql = f'DROP TABLE IF EXISTS {table_name};'
        try:
            cr = self.db_conn.cursor()
            cr.execute(sql)
        except sqlite3.Error as e:  # pragma: no cover
            Render.panel.failure(f'SQL drop db failed - SQL: "{sql}", Error: "{e}"')

    def db_insert_record(self, table_name: str, columns: list[str]):
        """Insert records into DB.

        A single row will all values as None so that values can be updated one at a
        time during parsing. The row and values will be used to determine permutations.

        Args:
            table_name: The DB table name.
            columns: The DB column names.
        """
        # sort and unique columns
        columns = sorted(set(columns))
        bindings = ', '.join(['?'] * len(columns))
        columns_string = ', '.join([f'''"{c.strip('"').strip("'")}"''' for c in columns])
        values = [None] * len(columns)
        sql = f'INSERT INTO {table_name} ({columns_string}) VALUES ({bindings})'
        try:
            cur = self.db_conn.cursor()
            cur.execute(sql, values)
        except sqlite3.OperationalError as ex:  # type: ignore
            Render.panel.failure(f'SQL insert failed - SQL: "{sql}", Error: "{ex}"')

    def db_update_record(self, table_name: str, column: str, value: bool | str | None):
        """Update a single column in the row-column create in db_insert_record.

        Args:
            table_name: The DB table name.
            column: The DB column names.
            value: The DB values to store (row data).
        """
        # escape any single quotes in value
        if isinstance(value, str):
            value = value.replace('\'', '\\')
        elif isinstance(value, bool):
            # core expects true/false so we convert bool value to string and lower
            value = str(value).lower()
        else:  # pragma: no cover
            # no other types can be used in a layout.json display clause
            return

        # only column defined in install.json can be updated
        if column in self.ij.model.param_names:
            # value should be wrapped in single quotes to be properly parsed
            sql = f'UPDATE {table_name} SET {column} = \'{value}\''
            try:
                cur = self.db_conn.cursor()
                cur.execute(sql)
            except sqlite3.OperationalError as ex:  # pragma: no cover
                Render.panel.failure(f'SQL update failed - SQL: "{sql}", Error: "{ex}"')

    def get_action_input_names(self, action: str) -> list[str]:
        """Return the input names for the provided action."""
        return [i.name for i in self.get_action_inputs(action)]

    def get_action_inputs(self, action: str) -> list[InputModel]:
        """Return the inputs for the provided action."""
        return self.action_configurations.get(action, {}).get('inputs', [])

    def get_action_output_names(self, action: str) -> list[str]:
        """Return the output names for the provided action."""
        return [i.name for i in self.get_action_outputs(action)]

    def get_action_outputs(self, action: str) -> list[OutputVariablesModel]:
        """Return the outputs for the provided action."""
        return self.action_configurations.get(action, {}).get('outputs', [])

    def get_input_applies_to_all(self, input_name: str) -> bool:
        """Return the outputs for the provided action."""
        for data in self.action_configurations.values():
            for d in data.get('inputs', []):
                if d.name == input_name:
                    break
            else:
                return False  # not found
        return True

    def extract_tc_action_clause(self, display_clause: str | None) -> str | None:
        """Extract the tc_action part of the display clause."""
        if display_clause is not None:
            # action_clause_extract_pattern = r'(tc_action\sin\s\([^\)]*\))'
            action_clause_extract_pattern = r'''(tc_action\sin\s\(.+?(?<='\)))'''
            _tc_action_clause = re.search(
                action_clause_extract_pattern, display_clause, re.IGNORECASE
            )
            if _tc_action_clause is not None:
                self.log.debug(f'action=extract-action-clause, display-clause={display_clause}')
                return _tc_action_clause.group(1)
        return None

    @property
    def inputs_ordered(self) -> Iterator[ParamsModel]:
        """Return inputs ordered properly.

        Layout based playbook Apps have the inputs ordered in
        the layout.json so that the display clause is guaranteed
        to have conditional defined before the clause is executed.
        """
        if self.lj.has_layout:
            for name in self.lj.model.param_names:
                param = self.ij.model.get_param(name)
                if param is not None:
                    yield param
        else:
            for param in self.ij.model.params:
                yield param

    # TODO: [low] improve this logic
    def init_permutations(self):
        """Process layout.json names/display to get all permutations of args."""
        if not all([self._input_permutations, self._output_permutations]):
            self._input_permutations = []
            self._output_permutations = []

            # create db for permutations testing
            self.db_create_table(self.input_table, self.ij.model.param_names)
            self.db_insert_record(self.input_table, self.ij.model.param_names)

            # only gen permutations if none have been generated previously
            self._gen_permutations()

            # drop database
            self.db_drop_table(self.input_table)

    def inputs_by_action(
        self, action: str, include_hidden: bool = True
    ) -> Iterator[dict[str, bool | ParamsModel]]:
        """Return all inputs for the provided action."""
        for ij_data in self._params_data:
            # get a display clause with just the tc_action condition
            display = self.extract_tc_action_clause(self.lj.model.get_param(ij_data.name).display)
            self.log.debug(f'action=input-by-action, input-name={ij_data.name}, display={display}')

            if ij_data.service_config:
                # inputs that are serviceConfig are not applicable for profiles
                continue

            if display is None:
                # when there is no display clause or the input is hidden,
                # then the input gets added to the AppBaseModel
                applies_to_all = True
            elif ij_data.hidden is True and include_hidden is True:
                applies_to_all = True
            elif self.validate_input_variable(ij_data.name, {'tc_action': action}, display):
                # each input will be checked for permutations if the App has layout and not hidden
                applies_to_all = False
            else:
                continue

            yield {'applies_to_all': applies_to_all, 'input': ij_data}

    def input_dict(self, permutation_id: int) -> dict:
        """Return all input permutation names for provided permutation id.

        {'tc_action': 'Append', 'input_strings': None, 'append_chars': None}

        Args:
            permutation_id: The index of the permutation input array.

        Returns:
            dict: A dict with key / value for each input for the provided permutation id.
        """
        input_dict = {}
        if self.lj.has_layout:
            for permutation in self.input_permutations[permutation_id]:
                input_dict.setdefault(permutation.name, permutation.value)
        return input_dict

    @cached_property
    def input_names(self) -> list[list[str]]:
        """Return all input permutation names for current App.

        Returns:
            list: List of Lists of input names.
        """
        input_names = []
        if self.lj.has_layout:
            for permutation in self.input_permutations:
                input_names.append([p.name for p in permutation])
        return input_names

    @property
    def input_permutations(self) -> list[list[InputModel]]:
        """Return all input permutations for current App.

        self._input_permutations is an array of permutations arrays.
        [[<perm obj #1], [<perm obj #2]]

        Returns:
            list: List of Lists of valid input permutations.
        """
        if not self._input_permutations and self.lj.has_layout:
            self.init_permutations()
        return self._input_permutations

    @property
    def output_permutations(self) -> list[list[OutputVariablesModel]]:
        """Return all output permutations for current App.

        Returns:
            list: List of Lists of valid outputs permutations.
        """
        if self._output_permutations is None:
            self.init_permutations()
        return self._output_permutations

    def outputs_by_action(self, action: str) -> Iterator[OutputVariablesModel]:
        """Return all inputs for the provided action."""
        yield from self.outputs_by_inputs({'tc_action': action})

    def outputs_by_inputs(self, inputs: dict[str, str]) -> Iterator[OutputVariablesModel]:
        """Return all output based on provided inputs

        Args:
            inputs: The args/inputs dict.

        Returns:
            list: List of Lists of valid outputs objects.
        """
        table = f'temp_{random.randint(100,999)}'  # nosec
        self.db_create_table(table, self.ij.model.param_names)
        self.db_insert_record(table, self.ij.model.param_names)

        for name, val in inputs.items():
            self.db_update_record(table, name, val)

        # iterate of InstallJsonModel -> PlaybookModel -> OutputVariablesModel
        if self.ij.model.playbook:
            for o in self.ij.model.playbook.output_variables:
                lj_output = self.lj.model.get_output(o.name)
                if isinstance(lj_output, NoneModel):  # pragma: no cover
                    # an output not listed in layout.json should always be shown
                    yield o
                elif self.validate_layout_display(table, lj_output.display):
                    # all other outputs must be validated
                    yield o

        # drop database
        self.db_drop_table(table)

    def permutations(self):
        """Process layout.json names/display to get all permutations of args."""
        if not self.lj.has_layout:  # pragma: no cover
            Render.panel.failure('Only Apps with a layout.json are supported.')

        if 'sqlite3' not in sys.modules:  # pragma: no cover
            Render.panel.failure(
                'The sqlite3 module needs to be built-in to Python for this feature.'
            )

        # create db for permutations testing
        self.db_create_table(self.input_table, self.ij.model.param_names)
        self.db_insert_record(self.input_table, self.ij.model.param_names)

        # only gen permutations if none have been generated previously
        if not all([self._input_permutations, self._output_permutations]):
            self._input_permutations = self._input_permutations or []
            self._output_permutations = self._output_permutations or []
            self._gen_permutations()

        # output permutations
        self.write_permutations_file()

    def validate_input_variable(
        self, input_name: str, inputs: dict, display: str | None = None
    ) -> bool:
        """Return True if the provided variables display where clause returns results.

        Args:
            input_name: The input variable name (e.g. tc_action).
            inputs: The current name/value dict.
            display: An optional display clause to validate. If not provided the
                display clause will be retrieved from the layout.json file.

        Returns:
            bool: True if the display value returns results.
        """
        # TODO: [low] re-evaluate this now that we only support layout.json Apps
        if not self.lj.has_layout or not inputs:  # pragma: no cover
            # always return true, even if current App doesn't have layouts
            return True

        table = f'temp_{random.randint(100,999)}'  # nosec
        self.db_create_table(table, self.ij.model.param_names)
        self.db_insert_record(table, self.ij.model.param_names)

        # APP-98 Added to cover the use case of interdependent variables in the layout.json.
        for name, param in self.ij.model.filter_params(_type='Boolean').items():
            self.db_update_record(table, name, param.default)

        for name, val in inputs.items():
            self.db_update_record(table, name, val)

        if display is None:
            lj_data = self.lj.model.get_param(input_name)
            if isinstance(lj_data, NoneModel):  # pragma: no cover
                # this shouldn't happen as all ij inputs must be in lj
                raise RuntimeError(f'The provided input {input_name} was not found in layout.json.')
            display = lj_data.display

        # check if provided variable meets display requirements
        valid = self.validate_layout_display(table, display)

        # cleanup temp table
        self.db_drop_table(table)

        return valid

    def validate_layout_display(self, table: str, display_condition: str | None) -> bool:
        """Check to see if the display condition passes.

        Args:
            table: The name of the DB table which hold the App data.
            display_condition: The "where" clause of the DB SQL statement.

        Returns:
            bool: True if the row count is greater than 0.
        """
        display = False
        if display_condition is None:
            display = True
        else:
            display_query = f'SELECT count(*) from {table} where {display_condition}'  # nosec
            try:
                cur = self.db_conn.cursor()
                cur.execute(display_query.replace('"', ''))
                rows = cur.fetchall()
                if rows[0][0] > 0:
                    display = True
            except sqlite3.Error as e:  # pragma: no cover
                Render.panel.failure(f'"{display_query}" query returned an error: ({e}).')
        return display

    def write_permutations_file(self):
        """Write permutations file."""
        _permutations = []
        for index, permutations in enumerate(self.input_permutations):
            _permutations.append(
                {'index': index, 'args': [{'name': p.name, 'value': p.value} for p in permutations]}
            )

        with self.fqfn.open(mode='w', encoding='utf-8') as fh:
            json.dump(_permutations, fh, indent=2, sort_keys=True)
        Render.panel.success('All permutations written to the "permutations.json" file.')



"""TcEx Framework Module"""
# standard library
import json
import logging
from mimetypes import MimeTypes
from typing import cast

# third-party
from requests import Response, Session
from requests.exceptions import RequestException

from ...app.playbook import Playbook
from ...input.model.advanced_request_model import AdvancedRequestModel
from ...logger.trace_logger import TraceLogger

# get tcex logger
_logger: TraceLogger = logging.getLogger(__name__.split('.', maxsplit=1)[0])  # type: ignore


class AdvancedRequest:
    """App Feature Advanced Request Module

    Args:
        inputs: The instance of App inputs.
        playbooks: An instance Playbooks.
        session: An instance of Requests Session object.
        output_prefix: The output prefix.
        timeout: The timeout value for the request.
    """

    def __init__(
        self,
        model: AdvancedRequestModel,
        playbook: Playbook,
        session: Session,
        output_prefix: str,
        timeout: int = 600,
    ):
        """Initialize instance properties."""
        self.model = model
        self.playbook = playbook
        self.output_prefix = output_prefix
        self.session = session
        self.timeout = timeout or 600

        # properties
        self.allow_redirects: bool = True
        self.data: bytes | dict | str | None = None
        self.headers: dict = {}
        self.log = _logger
        self.max_mb: int = 500
        self.mt = MimeTypes()
        self.params: dict = {}

    def configure_body(self):
        """Configure Body"""
        self.data = self.model.tc_adv_req_body
        if self.data is not None:
            # INT-1386
            try:
                self.data = self.data.encode('utf-8')  # type: ignore
            except AttributeError:
                pass  # Binary Data

        if self.model.tc_adv_req_urlencode_body:
            # the user has selected to urlencode the body, which indicates that
            # the body is a JSON string and should be converted to a dict
            self.data = cast(str, self.data)
            try:
                self.data = json.loads(self.data)
            except ValueError:  # pragma: no cover
                self.log.error('Failed loading body as JSON data.')

    def configure_headers(self):
        """Configure Headers

        [{
            "key": "User-Agent",
            "value": "TcEx MyApp: 1.0.0",
        }]
        """
        for header_data in self.model.tc_adv_req_headers or []:
            value = self.playbook.read.variable(header_data['value'])
            self.headers[str(header_data.get('key'))] = str(value)

    def configure_params(self):
        """Configure Params

        [{
            "count": "500",
            "page": "1",
        }]
        """
        for param_data in self.model.tc_adv_req_params or []:
            param = param_data.get('key')
            values = self.playbook.read.variable(param_data['value'])
            if not isinstance(values, list):
                values = [values]
            for value in values:
                if not value and self.model.tc_adv_req_exclude_null_params:
                    self.log.warning(
                        f'Query parameter {param} has a null/empty value '
                        'and will not be added to the request.'
                    )
                else:
                    self.params.setdefault(param, []).append(str(value))

    def request(self) -> Response | None:
        """Make the HTTP request."""
        if self.model.tc_adv_req_path is None:
            return None

        # configure body
        self.configure_body()

        # configure headers
        self.configure_headers()

        # configure params
        self.configure_params()

        # make http request
        try:
            response = self.session.request(
                allow_redirects=self.allow_redirects,
                data=self.data,
                headers=self.headers,
                method=self.model.tc_adv_req_http_method,
                params=self.params,
                timeout=self.timeout,
                url=self.model.tc_adv_req_path,
            )
        except RequestException as ex:  # pragma: no cover
            raise RuntimeError(f'Exception during request ({ex}).') from ex

        # write outputs as soon as they are available
        self.playbook.create.variable(
            f'{self.output_prefix}.request.headers', json.dumps(dict(response.headers)), 'String'
        )
        self.playbook.create.variable(
            f'{self.output_prefix}.request.ok', str(response.ok).lower(), 'String'
        )
        self.playbook.create.variable(
            f'{self.output_prefix}.request.reason', response.reason, 'String'
        )
        self.playbook.create.variable(
            f'{self.output_prefix}.request.status_code', str(response.status_code), 'String'
        )
        self.playbook.create.variable(
            f'{self.output_prefix}.request.url',
            response.request.url or self.model.tc_adv_req_path,
            'String',
        )

        # get response size
        response_bytes: int = len(response.content)
        response_mb: float = response_bytes / 1000000
        self.log.info(f'Response MB: {response_mb}')
        if response_mb > self.max_mb:  # pragma: no cover
            raise RuntimeError('Download was larger than maximum supported 500 MB.')

        # write content after size validation
        self.playbook.create.variable(
            f'{self.output_prefix}.request.content', response.text, 'String'
        )
        self.playbook.create.variable(
            f'{self.output_prefix}.request.content.binary', response.content, 'Binary'
        )

        # fail if fail_on_error is selected and not ok
        if self.model.tc_adv_req_fail_on_error and not response.ok:
            raise RuntimeError(f'Failed for status ({response.status_code})')

        return response



"""TcEx Framework Module"""
# standard library
from enum import Enum

# first-party
from tcex.api.tc.v3.api_endpoints import ApiEndpoints
from tcex.api.tc.v3.filter_abc import FilterABC
from tcex.api.tc.v3.tql.tql_type import TqlType


class WorkflowTemplateFilter(FilterABC):
    """Filter Object for WorkflowTemplates"""

    @property
    def _api_endpoint(self) -> str:
        """Return the API endpoint."""
        return ApiEndpoints.WORKFLOW_TEMPLATES.value

    def active(self, operator: Enum, active: bool):
        """Filter Active based on **active** keyword.

        Args:
            operator: The operator enum for the filter.
            active: The active status of this template.
        """
        self._tql.add_filter('active', operator, active, TqlType.BOOLEAN)

    def description(self, operator: Enum, description: list | str):
        """Filter Description based on **description** keyword.

        Args:
            operator: The operator enum for the filter.
            description: The description of this template.
        """
        if isinstance(description, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('description', operator, description, TqlType.STRING)

    def id(self, operator: Enum, id: int | list):  # pylint: disable=redefined-builtin
        """Filter ID based on **id** keyword.

        Args:
            operator: The operator enum for the filter.
            id: The ID of the template.
        """
        if isinstance(id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('id', operator, id, TqlType.INTEGER)

    def name(self, operator: Enum, name: list | str):
        """Filter Name based on **name** keyword.

        Args:
            operator: The operator enum for the filter.
            name: The name of this template.
        """
        if isinstance(name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('name', operator, name, TqlType.STRING)

    def owner(self, operator: Enum, owner: int | list):
        """Filter Owner ID based on **owner** keyword.

        Args:
            operator: The operator enum for the filter.
            owner: The Owner ID for the template.
        """
        if isinstance(owner, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('owner', operator, owner, TqlType.INTEGER)

    def owner_name(self, operator: Enum, owner_name: list | str):
        """Filter Owner Name based on **ownerName** keyword.

        Args:
            operator: The operator enum for the filter.
            owner_name: The owner name for the template.
        """
        if isinstance(owner_name, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('ownerName', operator, owner_name, TqlType.STRING)

    def target_id(self, operator: Enum, target_id: int | list):
        """Filter Target ID based on **targetId** keyword.

        Args:
            operator: The operator enum for the filter.
            target_id: The assigned user or group ID for the template.
        """
        if isinstance(target_id, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('targetId', operator, target_id, TqlType.INTEGER)

    def target_type(self, operator: Enum, target_type: list | str):
        """Filter Target Type based on **targetType** keyword.

        Args:
            operator: The operator enum for the filter.
            target_type: The target type for this template (either User or Group).
        """
        if isinstance(target_type, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('targetType', operator, target_type, TqlType.STRING)

    def version(self, operator: Enum, version: int | list):
        """Filter Version based on **version** keyword.

        Args:
            operator: The operator enum for the filter.
            version: The version of this template.
        """
        if isinstance(version, list) and operator not in self.list_types:
            raise RuntimeError(
                'Operator must be CONTAINS, NOT_CONTAINS, IN'
                'or NOT_IN when filtering on a list of values.'
            )

        self._tql.add_filter('version', operator, version, TqlType.INTEGER)



"""TcEx Framework Module"""

# first-party
from tcex.api.tc.v3._gen.model._filter_model import FilterModel
from tcex.api.tc.v3._gen.model._property_model import PropertyModel

__all__ = ['FilterModel', 'PropertyModel']



"""TcEx Framework Module"""
# pylint: disable=no-member,no-self-argument,wrong-import-position
# third-party
from pydantic import BaseModel, Extra, Field, PrivateAttr, validator

# first-party
from tcex.api.tc.v3.v3_model_abc import V3ModelABC
from tcex.util import Util


class UserGroupModel(
    V3ModelABC,
    alias_generator=Util().snake_to_camel,
    extra=Extra.allow,
    title='UserGroup Model',
    validate_assignment=True,
):
    """User_Group Model"""

    _associated_type = PrivateAttr(False)
    _cm_type = PrivateAttr(False)
    _shared_type = PrivateAttr(False)
    _staged = PrivateAttr(False)

    description: str | None = Field(
        None,
        allow_mutation=False,
        description='The **description** for the User_Group.',
        read_only=True,
        title='description',
    )
    id: int | None = Field(
        None,
        description='The ID of the item.',
        read_only=True,
        title='id',
    )
    name: str | None = Field(
        None,
        allow_mutation=False,
        description='The **name** for the User_Group.',
        read_only=True,
        title='name',
    )
    users: 'UsersModel' = Field(
        None,
        allow_mutation=False,
        description='The **users** for the User_Group.',
        read_only=True,
        title='users',
    )

    @validator('users', always=True, pre=True)
    def _validate_users(cls, v):
        if not v:
            return UsersModel()  # type: ignore
        return v


class UserGroupDataModel(
    BaseModel,
    title='UserGroup Data Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """User_Groups Data Model"""

    data: list[UserGroupModel] | None = Field(
        [],
        description='The data for the UserGroups.',
        methods=['POST', 'PUT'],
        title='data',
    )


class UserGroupsModel(
    BaseModel,
    title='UserGroups Model',
    alias_generator=Util().snake_to_camel,
    validate_assignment=True,
):
    """User_Groups Model"""

    _mode_support = PrivateAttr(False)

    data: list[UserGroupModel] | None = Field(
        [],
        description='The data for the UserGroups.',
        methods=['POST', 'PUT'],
        title='data',
    )
    mode: str = Field(
        'append',
        description='The PUT mode for nested objects (append, delete, replace). Default: append',
        methods=['POST', 'PUT'],
        title='append',
    )


# first-party
from tcex.api.tc.v3.security.users.user_model import UsersModel

# add forward references
UserGroupDataModel.update_forward_refs()
UserGroupModel.update_forward_refs()
UserGroupsModel.update_forward_refs()



"""TcEx Framework Module"""
# standard library
import logging

# third-party
from redis import Redis

from ...input.field_type.sensitive import Sensitive
from ...pleb.cached_property import cached_property
from ...pleb.scoped_property import scoped_property
from ...requests_tc.tc_session import TcSession
from .key_value_api import KeyValueApi
from .key_value_mock import KeyValueMock
from .key_value_redis import KeyValueRedis
from .redis_client import RedisClient

# get logger
_logger = logging.getLogger(__name__.split('.', maxsplit=1)[0])


class KeyValueStore:
    """TcEx Module"""

    def __init__(
        self,
        session_tc: TcSession,
        tc_kvstore_host: str,
        tc_kvstore_port: int,
        tc_kvstore_type: str,
        tc_kvstore_pass: Sensitive | None = None,
        tc_kvstore_user: str | None = None,
    ):
        """Initialize the class properties."""
        self.session_tc = session_tc
        self.tc_kvstore_host = tc_kvstore_host
        self.tc_kvstore_pass = tc_kvstore_pass
        self.tc_kvstore_port = tc_kvstore_port
        self.tc_kvstore_type = tc_kvstore_type
        self.tc_kvstore_user = tc_kvstore_user

        # properties
        self.log = _logger

    @scoped_property
    def client(self) -> KeyValueApi | KeyValueMock | KeyValueRedis:
        """Return the correct KV store for this execution.

        The TCKeyValueAPI KV store is limited to two operations (create and read),
        while the Redis kvstore wraps a few other Redis methods.
        """
        if self.tc_kvstore_type == 'Redis':
            # submodule doesn't have scoped_property decorator, so resolution of type doesn't work
            return KeyValueRedis(self.redis_client)  # type: ignore

        if self.tc_kvstore_type == 'TCKeyValueAPI':
            return KeyValueApi(self.session_tc)  # pylint: disable=no-member

        if self.tc_kvstore_type == 'Mock':
            self.log.warning(
                'Using mock key-value store. '
                'This should *never* happen when running in-platform.'
            )
            return KeyValueMock()

        raise RuntimeError(f'Invalid KV Store Type: ({self.tc_kvstore_type})')

    @cached_property
    def client_kvr(self) -> KeyValueRedis:
        """Return the Redis KV store client.

        This property should only be used when the KV store type is Redis.
        """
        return KeyValueRedis(self.redis_client)

    @staticmethod
    def get_redis_client(
        host: str, port: int, db: int = 0, blocking_pool: bool = False, **kwargs
    ) -> Redis:
        """Return a *new* instance of Redis client.

        For a full list of kwargs see https://redis-py.readthedocs.io/en/latest/#redis.Connection.

        Args:
            host: The REDIS host. Defaults to localhost.
            port: The REDIS port. Defaults to 6379.
            db: The REDIS db. Defaults to 0.
            blocking_pool: Use BlockingConnectionPool instead of ConnectionPool.
            **kwargs: Additional keyword arguments.

        Keyword Args:
            errors (str): The REDIS errors policy (e.g. strict).
            max_connections (int): The maximum number of connections to REDIS.
            password (Sensitive): The REDIS password.
            socket_timeout (int): The REDIS socket timeout.
            timeout (int): The REDIS Blocking Connection Pool timeout value.
            username (str): The REDIS username.
        """
        # get value from Sensitive value before passing to Redis
        password = kwargs.get('password')
        kwargs['password'] = password.value if isinstance(password, Sensitive) else password
        return RedisClient(
            host=host, port=port, db=db, blocking_pool=blocking_pool, **kwargs
        ).client

    @scoped_property
    def redis_client(self) -> Redis:
        """Return redis client instance configure for Playbook/Service Apps."""
        return self.get_redis_client(
            host=self.tc_kvstore_host,
            port=self.tc_kvstore_port,
            db=0,
            username=self.tc_kvstore_user,
            password=self.tc_kvstore_pass,
        )



"""TcEx Framework Module"""
# standard library
from copy import deepcopy
from threading import Lock

from .key_value_abc import KeyValueABC


class KeyValueMock(KeyValueABC):
    """TcEx Key Value Mock Module.

    Purely in-memory implementation of the KeyValueABC for local testing only.
    """

    data = {}

    def __init__(self):
        """Initialize the Class properties."""
        self.lock = Lock()

        # properties
        self.kv_type = 'mock'

    def create(self, context: str, key: str, value: bytes | str) -> int:
        """Create key/value pair.

        Args:
            context: A specific context for the create.
            key (str): The field name (key) for the kv pair in Redis.
            value (any): The value for the kv pair in Redis.

        Returns:
            str: The response from Redis.
        """
        with self.lock:
            self.data.setdefault(context, {})[key] = value
            return 1

    def read(self, context: str, key: str) -> bytes | str | None:
        """Read data for the provided key.

        Args:
            context: A specific context for the create.
            key: The field name (key) for the kv pair in Redis.

        Returns:
            str: The response data from Redis.
        """
        with self.lock:
            return self.data.get(context, {}).get(key)

    def get_all(self, context: str | None) -> dict[str, bytes | str | None]:
        """Return the contents for a given context.

        Args:
            context: the context to return
        """
        if context is not None:
            return deepcopy(self.data.get(context, {}))

        return deepcopy(self.data)
